from ._dearpypixl import (
    common as typing,
)
from ._dearpypixl.api import (
    Application as Application,
    Viewport as Viewport,
    Runtime as Runtime,
    Registry as Registry,
)
from .items import *



import types
import enum
import functools
import itertools  # type: ignore
import os
from dearpygui import dearpygui
from ._dearpypixl.common import (
    Any,
    Item,
    overload,
)
from . import items, color, style
from ._dearpypixl import api, interface, tools
from ._dearpypixl.api import Registry as RegistryAPI, Item as ItemAPI




# TODO: re-implement ez-theme

Theme = items.Theme



class FontRangeHint(enum.Enum):
    # XXX I suspect that the internal structure accepting characters expects
    # unsigned shorts. Characters exceeding a value of 65535 (some below)
    # are seemingly ignored.

    DEFAULT = tuple(range(0x0000, 0x00ff))  # basic Latin + Latin, 0x1 Supplement
    LATIN = tuple(itertools.chain(
        range(0x0100 , 0x017F  + 1),     # Latin Extended A
        range(0x0180 , 0x024F  + 1),     # Latin Extended B
        range(0x0250 , 0x02AF  + 1),     # IPA Extensions
        range(0x02B0 , 0x02FF  + 1),     # Spacing Modifier Letters
        range(0x1D00 , 0x1D7F  + 1),     # Phonetic Extensions
        range(0x1D80 , 0x1DBF  + 1),     # Phonetic Extensions Supplement
        range(0x1E00 , 0x1EFF  + 1),     # Latin Extended Additional
        range(0x2070 , 0x209F  + 1),     # Superscripts and Subscripts
        range(0x2100 , 0x214F  + 1),     # Letterlike Symbols
        range(0x2150 , 0x218F  + 1),     # Number Forms
        range(0x2C60 , 0x2C7F  + 1),     # Latin Extended C
        range(0xA720 , 0xA7FF  + 1),     # Latin Extended D
        range(0xAB30 , 0xAB6F  + 1),     # Latin Extended E
        range(0xFB00 , 0xFB4F  + 1),     # Alphabetic Presentation Forms (Latin ligatures)
        range(0xFF00 , 0xFFEF  + 1),     # Halfwidth and Fullwidth Forms
        range(0x10780, 0x107BF + 1),     # Latin Extended F
        range(0x1DF00, 0x1DFFF + 1),     # Latin Extended G
    ))
    JAPANESE = tuple(itertools.chain(
        # Hiragana
        range(0x3041, 0x3096 + 1),
        # Katakana
        range(0x30a0, 0x30ff + 1),   # full-width
        range(0xff5f, 0xff9f + 1),   # half-width (+punctuation)
        # Kanji
        range(0x3400, 0x4db5 + 1),
        range(0x4e00, 0x9fcb + 1),
        range(0xF900, 0xFA6A + 1),
        range(0x2e80, 0x2fd5 + 1),  #  radicals
        range(0x3000, 0x303f + 1),  # |
        range(0x31f0, 0x31ff + 1),  # |- symbols, punctuation, & misc characters
        range(0x3220, 0x3243 + 1),  # |
        range(0x3280, 0x337F + 1),  # |
    ))
    KOREAN = tuple(itertools.chain(
        range(0x2000, 0x4000 + 1),
        range(0x5B00, 0x5F00 + 1),
        range(0x7B00, 0x7D00 + 1),
        range(0xA000, 0xA100 + 1),
        range(0xA700, 0xA900 + 1),
        range(0xB600, 0xB700 + 1),
        range(0x2010, 0x2011 + 1),
        range(0x2014, 0x2015 + 1),
        range(0x2018, 0x2019 + 1),
        range(0x201C, 0x201D + 1),
        range(0x2020, 0x2021 + 1),
        range(0x2025, 0x2026 + 1),
        range(0x2032, 0x2033 + 1),
        range(0x3001, 0x3003 + 1),
        range(0x3008, 0x3011 + 1),
        range(0x3014, 0x3015 + 1),
        range(0x3141, 0x3142 + 1),
        range(0x3147, 0x3148 + 1),
        range(0x314A, 0x314E + 1),
        range(0xAC00, 0xD7A3 + 1),
        range(0xFF01, 0xFF03 + 1),
        range(0xFF05, 0xFF0A + 1),
        range(0xFF0C, 0xFF0F + 1),
        range(0xFF1A, 0xFF1B + 1),
        range(0xFF1F, 0xFF20 + 1),
        range(0xFF3B, 0xFF3D + 1),
        (0xBF00, 0x2030, 0x203B, 0x203E,
         0x20AC, 0x301C, 0x30FB, 0x3131,
         0x3134, 0x3137, 0x3139, 0x3145,
         0xFF3F, 0xFF5B, 0xFF5D),
    ))
    CHINESE = tuple(itertools.chain(
        range(0x4E00, 0x9FFF + 1),  # CJK Unified Ideographs
        range(0x2f00, 0x2fdf + 1),  # CJK/Kangxi Radicals
        range(0x2e80, 0x2EFF + 1),  # CJK Radicals Supplement
        range(0x3000, 0x303F + 1),  # CJK symbols & punctuation
    ))
    CHINESE_SIMPLIFIED_COMMON = CHINESE
    CHINESE_FULL = tuple(itertools.chain(
        range(0x3400 , 0x4DBF  + 1),  # CJK Unified Ideographs Extension A      (rare)
        range(0x20000, 0x2A6DF + 1),  # CJK Unified Ideographs Extension B      (rare, historic)
        range(0x2A700, 0x2B73F + 1),  # CJK Unified Ideographs Extension C      (rare, historic)
        range(0x2B740, 0x2B81F + 1),  # CJK Unified Ideographs Extension D      (uncommon, some in current use)
        range(0x2B820, 0x2CEAF + 1),  # CJK Unified Ideographs Extension E      (rare, historic)
        range(0x2CEB0, 0x2EBEF + 1),  # CJK Unified Ideographs Extension F      (rare, historic)
        range(0x30000, 0x3134F + 1),  # CJK Unified Ideographs Extension G      (rare, historic)
        range(0x31350, 0x323AF + 1),  # CJK Unified Ideographs Extension H      (rare, historic)
        range(0xF900 , 0xFAFF  + 1),  # CJK Compatibility Ideographs            (duplicates, unifiable variants, corporate characters)
        range(0x2F800, 0x2FA1F + 1),  # CJK Compatibility Ideographs Supplement (unifiable variants)
    ))
    CYRILLIC = tuple(itertools.chain(
        range(0x0400 , 0x04FF  + 1),  # basic (256 characters)
        range(0x0500 , 0x052F  + 1),  # Supplement
        range(0x2DE0 , 0x2DFF  + 1),  # Ext-A
        range(0xA640 , 0xA69F  + 1),  # Ext-B
        range(0x1C80 , 0x1C8F  + 1),  # Ext-C
        range(0x1E030, 0x1E08F + 1),  # Ext-D
        range(0xFE2E , 0xFE2F  + 1),  # Combining Half Marks
        range(0x1D2B , 0x1D78  + 1),  # Phonetic Extensions
    ))
    THAI = tuple(range(0x0e00, 0x0e7f + 1))
    VIETNAMESE = tuple(itertools.chain(
        range(0x0020, 0x002F + 1),
        range(0x0030, 0x0039 + 1),
        range(0x003A, 0x0040 + 1),
        range(0x0041, 0x005A + 1),
        range(0x005B, 0x0060 + 1),
        range(0x0061, 0x007A + 1),
        range(0x007B, 0x007E + 1),
        range(0x00C0, 0x00C3 + 1),
        range(0x00C8, 0x00CA + 1),
        range(0x00CC, 0x00CD + 1),
        range(0x00D2, 0x00D5 + 1),
        range(0x00D9, 0x00DA + 1),
        range(0x00E0, 0x00E3 + 1),
        range(0x00E8, 0x00EA + 1),
        range(0x00EC, 0x00ED + 1),
        range(0x00F2, 0x00F5 + 1),
        range(0x00F9, 0x00FA + 1),
        range(0x0102, 0x0103 + 1),
        range(0x0110, 0x0111 + 1),
        range(0x0128, 0x0129 + 1),
        range(0x0168, 0x0169 + 1),
        range(0x01A0, 0x01B0 + 1),
        range(0x1EA0, 0x1EF9 + 1),
        range(0x02C6, 0x0323 + 1),
        (0x00DD, 0x00FD, 0x00D0),
    ))
    ALL = tuple(range(256, 0xffff + 1))

    @classmethod
    def _missing_(cls, value: Any) -> 'FontRangeHint | None':
        return _MVFRH_TO_PXFRH.get(value, None)


_MVFRH_TO_PXFRH: dict[int, FontRangeHint] = {  # helper for `FontRangeHint._missing_`
    dearpygui.mvFontRangeHint_Default                  : FontRangeHint.DEFAULT,
    dearpygui.mvFontRangeHint_Japanese                 : FontRangeHint.JAPANESE,
    dearpygui.mvFontRangeHint_Korean                   : FontRangeHint.KOREAN,
    dearpygui.mvFontRangeHint_Chinese_Full             : FontRangeHint.CHINESE_FULL,
    dearpygui.mvFontRangeHint_Chinese_Simplified_Common: FontRangeHint.CHINESE_SIMPLIFIED_COMMON,
    dearpygui.mvFontRangeHint_Cyrillic                 : FontRangeHint.CYRILLIC,
    dearpygui.mvFontRangeHint_Thai                     : FontRangeHint.THAI,
    dearpygui.mvFontRangeHint_Vietnamese               : FontRangeHint.VIETNAMESE,
}





class SizedFont(items.mvFontRegistry):
    """A DearPyGui font registry interface that manages font items of different sizes
    of a specific font.

    A high-level perspective of `pxFont` -- it acts less like a font registry item and more
    like an individual font item.
        >>> font_fpath: str = ""  # "ProggyClean-Black.ttf", etc.
        >>> # include a default size and other sizes
        >>> pc_fonts = pxFonts(font_fpath, 12.0, 14.0, 18.0, default_size=16.0)
        >>> pc_fonts.add_characters(0x2126, 0xfb26)  # 'ohm', 'half-life'
        >>>
        >>> wndw1 = dpg.add_window(label="\\u2126")
        >>> wndw2 = dpg.add_window(label="\\ufb26")
        >>> pc_fonts.bind(wndw1, wndw2)  # multiple items per call

    Want to change the font size? It's your choice of an assignment or method call.
        >>> pc_fonts.size = 12         # rounded down -- sets as 12.0
        >>> pc_fonts.size = 15.08      # rounded down -- sets as 15.0
        >>> pc_fonts.set_size(28.04)   # rounded down -- sets as 28.0

    Since DearPyGui loads glyphs from any new font into a single texture, a font's `size`
    can't actually be updated. Instead, the registry manages individual font items of
    different sizes. When setting the size, it will search its cache for a font of *size*
    and re-bind all bound items to that font item. If a font of of *size* is not cached,
    the registry will create one (using the font file passed on initialization) and cache it
    for future use. Calling the instance and passing a *size* argument will return the a
    `mvFont` interface for the underlying font item of that size.

    Management of additional characters can be done through various `.add_char*` and `.del_char*`
    methods. Adding or removing additional characters, remaps, or range hints will update all
    underlying font items accordingly. A read-only view of additional and remapped characters
    can be accessed through the `.characters` and `.remapped_characters` properties respectively.
    As an implementation detail, `pxFont` does not add range hints or character ranges using
    DearPyGui's `add_font_range` or `add_font_range_hint` functions; character ranges are
    calculated and added to font items as additional characters, while font range hints -- members
    of the 'FontRangeHint' enumeration containing pre-defined character sets -- are handled
    similarly.

    NOTE: Fonts generally do not support all codepoints/glyphs, and many use different
    codepoint-to-glyph mappings. If a glyph/codepoint does not appear in DearPyGui's
    font manager, it's likely that the font does not have a glyph mapped to that codepoint
    (i.e. it's unsupported).
    """
    DEFAULT                   = FontRangeHint.DEFAULT
    LATIN                     = FontRangeHint.LATIN
    JAPANESE                  = FontRangeHint.JAPANESE
    KOREAN                    = FontRangeHint.KOREAN
    CHINESE                   = FontRangeHint.CHINESE                    # |- these are the same
    CHINESE_SIMPLIFIED_COMMON = FontRangeHint.CHINESE_SIMPLIFIED_COMMON  # |
    CHINESE_FULL              = FontRangeHint.CHINESE_FULL
    CYRILLIC                  = FontRangeHint.CYRILLIC
    THAI                      = FontRangeHint.THAI
    VIETNAMESE                = FontRangeHint.VIETNAMESE
    ALL                       = FontRangeHint.ALL

    __glbl_lock  = tools.Lock()
    __glbl_binds = {}

    _font_cache : dict[float, Item]
    _bound_items: frozenset[Item]
    _characters : frozenset[int]
    _char_remaps: dict[int, int]

    def __init__(self, file: str, *sizes: float, label: str = "", default_size: float = 16.0, **kwargs) -> None:
        """Args:
            * file : Path (string) pointing to a .ttf or .otf font file.

            * sizes: Sizes (other than *default_size*) to pre-load.

            * default_size: The initial "active size" for the registry. Defaults to 16.0.
        """
        file = str(file)
        self._filename = file.split(os.sep)[-1]
        super().__init__(label=label or f"pxFont [{self._filename}]", **kwargs)
        ItemAPI.set_alias(self, f"{file}##{self.tag}")

        self._font_cache   = {}
        self._bound_items  = frozenset()
        self._characters   = frozenset()
        self._char_remaps  = {}
        self._active_size  = default_size
        self._font_factory = functools.partial(
            dearpygui.add_font,
            file,
            user_data=self,
            parent=self,
            use_internal_label=True,
        )
        self._get_font_of_size(default_size)
        self.add_sizes(*sizes)

    def __call__(self, size: float) -> items.mvFont:
        return self._get_font_of_size(size)

    def _get_font_of_size(self, size: float) -> items.mvFont:
        try:
            size = float(f"{size:1f}")
        except TypeError:
            raise TypeError(
                f"expected float for `size` (got {type(size).__qualname__!r})."
            ) from None

        if font:=self._font_cache.get(size, None):
            return items.mvFont.new(font)

        font = self._font_factory(
            size,  # type: ignore
            label=f"{self._filename} [{size}]",
        )
        self._font_cache[size] = font
        self._upd_font_details(font)
        return items.mvFont.new(font)

    def _upd_font_details(self, font_item: Item = 0) -> None:
        """Update one or all fonts in the registry."""
        fonts = self._font_cache.values() if not font_item else (font_item,)
        chars = tuple(self.characters)
        for font in fonts:
            RegistryAPI.delete_item(font, children_only=True, slot=1)
            dearpygui.add_font_chars(chars, parent=font)

    def bind(self, *items: Item, app_level: bool = False) -> None:
        """Bind items to the registry and/or set the registry to apply the application-level
        font. They will reflect a font of `self.file` and `self.size`.

        Args:
            * item: Identifiers of items to bind.

            * app_level: If True, the application default font will always be set to a font of
            `self.file` and `self.size`. Default is False.
        """
        self_id = self.tag
        font_id = self._get_font_of_size(self._active_size)
        bound_items   = set(self._bound_items)
        missing_items = set()
        with self.__glbl_lock:
            try:
                for item in items:
                    try:
                        ItemAPI.set_font(item, font_id)
                    except SystemError:
                        if not RegistryAPI.item_exists(item):
                            missing_items.add(item)
                            continue
                        raise
                    bound_items.add(item)
                if app_level:
                    api.Application.set_font(font_id)
            finally:
                self.__glbl_binds.update({item: self_id for item in bound_items})
                for item in missing_items:
                    if item in self.__glbl_binds:
                        del self.__glbl_binds[item]
                self._bound_items = frozenset(bound_items.difference(missing_items))

    def unbind(self, *items: Item, app_level: bool = False) -> None:
        with self.__glbl_lock:
            for item in items:
                try:
                    ItemAPI.set_font(item, 0)
                except SystemError:
                    pass
            if app_level:
                api.Application.set_font(0)
            for item in items:
                if item in self.__glbl_binds:
                    del self.__glbl_binds[item]
            self._bound_items = self._bound_items.difference(items)

    @property
    def file(self) -> str:
        """[get] Return the file used for fonts created and parented by the
        registry.
        """
        return self.alias.split("##")[0]  # type: ignore


    @property
    def characters(self) -> frozenset[int]:
        """[get] Return all additional font characters in the registry."""
        return self._characters

    def add_characters(self, *characters: int) -> None:
        """Add additional characters to the registry.

        Args:
            * characters: Codepoints (unicode) to add.
        """
        self._characters = self._characters.union(characters)
        self._upd_font_details()

    def del_characters(self, *characters: int) -> None:
        """Remove characters formerly added from the registry.

        Characters that are not additonally added are ignored.
        """
        self._characters = self._characters.difference(characters)
        self._upd_font_details()

    def add_char_range(self, start: int, stop: int) -> None:
        """Add a range of characters to the registry."""
        # XXX `add_font_range` "items" are not used because;
        #   - glyphs added this way can be accessed only by viewing them in the
        #   font manager (cannot query DPG)
        #   - it reduces item/uuid bloat
        #   - simplifies pxFont structure
        self.add_characters(*(c for c in range(start, stop)))

    def del_char_range(self, start: int, stop: int) -> None:
        """Remove a range of characters formerly added to the registry.

        Characters that are not additonally added are ignored.
        """
        self.del_characters(*(c for c in range(start, stop)))

    def add_char_hint(self, range_hint: int | FontRangeHint) -> None:
        """Add characters to the registry from a pre-defined arrangement.

        Args:
            * range_hint: A member of `FontRangeHint` or a DearPyGui "mvFontRangeHint_*"
            constant.
        """
        # XXX `add_font_range_hint` "items" are not used because;
        #   - glyphs added from those hints do not show in the font manager
        #   - it reduces item/uuid bloat
        #   - simplifies pxFont structure
        try:
            characters = FontRangeHint(range_hint).value
        except TypeError:
            raise TypeError(
                f"expected `int`, `FontRangeHint` member (got {type(range_hint).__qualname__!r})."
            ) from None
        self.add_characters(*characters)  # type: ignore

    def del_char_hint(self, range_hint: int | FontRangeHint) -> None:
        """Remove characters formerly added to the registry from a pre-defined
        arrangement.

        Args:
            * range_hint: A member of `FontRangeHint` or DearPyGui "mvFontRangeHint_*"
            integer constant.
        """
        try:
            characters = FontRangeHint(range_hint).value
        except TypeError:
            raise TypeError(
                f"expected `int`, `FontRangeHint` member (got {type(range_hint).__qualname__!r})."
            ) from None
        self.del_characters(*characters)  # type: ignore


    @property
    def remapped_characters(self) -> types.MappingProxyType[int, int]:
        """[get] Return a mapping of remapped characters."""
        return types.MappingProxyType(self._char_remaps)

    def add_remapped_char(self, character: int, new_character: int) -> None:
        """Remap an existing character to a new one."""
        self._char_remaps[character] = new_character
        self._upd_font_details()

    def del_remapped_char(self, *characters: int) -> None:
        """Restore the original mapping of a remapped character.

        Ignores characters that are not remapped.
        """
        for c in characters:
            self._char_remaps.pop(c, None)
        self._upd_font_details()


    @property
    def size(self) -> float:
        """[get] Return the size currently in use by items bound to the registry."""
        return self._active_size
    @size.setter
    def size(self, value: float):
        """[set] Set the active font size for items bound to the registry."""
        self.set_size(value)

    def set_size(self, size: float) -> None:
        """Set the active font size for items bound to the registry."""
        self._active_size = size
        self.bind(*self._bound_items, app_level=True if api.Application.get_font() == self else False)

    @property
    def sizes(self) -> tuple[float, ...]:
        """[get] Return all font sizes loaded in the registry."""
        return tuple(sorted(self._font_cache))

    def add_sizes(self, *sizes: float) -> None:
        """Ensure that font items of *sizes* exist in the registry. Useful for
        "pre-loading" font items.

        This method exists for convenience -- the registry creates font items
        automatically as necessary.
        """
        for size in sizes:
            self._get_font_of_size(size)

    @overload
    def add_size_range(self, stop: float, /) -> None: ...
    @overload
    def add_size_range(self, start: float, stop: float, step: float = 1.0) -> None: ...
    def add_size_range(self, start: float, stop: float | None = None, step: float = 1.0) -> None:
        """Create missing font items from a range of sizes. Useful for
        "pre-loading" font items.

        This method exists for convenience -- the registry creates font items
        automatically as necessary."""
        if stop is None:
            rng = range(start)  # type: ignore
        else:
            i = start
            rng = []
            while i < stop:
                rng.append(i)
                i += step
        self.add_sizes(*rng)



"""Items for emulating consoles and other streams."""
import sys   # type: ignore
import types
import codeop
import traceback
import functools
import threading
import itertools  # type: ignore
import contextlib
import types
from dearpygui import dearpygui
from dearpypixl._dearpypixl.common import Item, Color
from . import items, color, style
from ._dearpypixl import api, interface, tools
from ._dearpypixl.api import Item as ItemAPI
from ._dearpypixl.common import (
    Item,
    Any,
    Sequence,
    ItemCommand,
    Callable,
    Self,
    Literal,
    SupportsIndex,
    Generator,
    TypeVar,
    ParamSpec,
    Method,
    override,
)


_T = TypeVar("_T")
_P = ParamSpec("_P")





def _null_item_processing(self: 'FileStream', obj: _T) -> _T:
    return obj

def _null_value_processing(self: 'FileStream', obj: _T, value: Any) -> _T:
    return value

def _item_generator(item_factory: Callable[_P, _T], *args: _P.args, **kwargs: _P.kwargs) -> Generator[_T, None, None]:
    while True:
        yield item_factory(*args, **kwargs)


class FileStream(interface.SupportsValueArray[str], items.mvValueRegistry):
    """A value registry that acts like pseudo file stream or buffer. It
    represents itself as a living multiline string split at newlines.
        >>> f = FileStream()
        >>> f.value = "foo\\nbar"
        >>> f.value
        ['f', 'o', 'o', '\\n', 'b', 'a', 'r']
        >>> f[0] = 'b'
        >>> f.value
        ['b', 'o', 'o', '\\n', 'b', 'a', 'r']
        >>> f[-1] = "FOO"
        >>> f.value
        ['b', 'o', 'o', '\\n', 'b', 'a', 'FOO']
        >>> f.value = ["foobar"]
        >>> f.value
        ["foobar"]

    Writing to the stream appends the value written:
        >>> f.write("helloworld")
        >>> f.value
        ['foobar', 'helloworld']

    The stream will always use the string representation of any object used
    as a value or values. When setting its' value as a whole:
        >>> f.value = ['a', 'b', None, 'd']
        >>> f.value
        ['a', 'b', 'None', 'd']

    Setting part of the value at a specific line number/index:
        >>> f[1] = None
        >>> f.value
        ['a', 'None', 'None', 'd']

    The string representation of the stream is the assembled value as if
    joined by newlines:
        >>> str(f)
        'a\\nNone\\nNone\\nd\\n'


    Individual 'mvStringValue' items are created to store each value added to
    the registry. These items can be used as a source for other items. Iterating
    through the stream will yield the identifier of the 'mvStringValue' and its
    value. Alternatively, you can get the identifier of a value item at a specific
    line number by indexing the stream's children (slot 1).

    The stream accepts three optional callbacks during initialization. This
    makes it easier to interface with the stream when it performs notable
    operations. For example, when making a primitive file display:
        >>> with dpg.window() as file_window: ...
        >>>
        >>> def add_line(buffer: FileStream, value_item: int | str):
        ...     # Add a new line to the display when a value is added to
        ...     # the stream. If the value is updated at this line, the
        ...     # text will display accordingly.
        ...     # Also, store the text item id within the value item for
        ...     # `del_line`.
        ...     text_item = dpg.add_text(parent=file_window, source=value_item)
        >>>
        >>> def del_line(buffer: FileStream, value_item: int | str):
        ...     # Destroy the associated text item when the value item is
        ...     # deleted.
        ...     dpg.delete_item(dpg.get_item_configuration(value_item)["user_data"])
        >>>
        >>> f = FileStream(
        ...     ["f", "o", "o", "bar"],
        ...     add_item_callback=add_line,
        ...     del_item_callback=del_line,
        ... )
    In the above example, writing to `f` will display content written in
    `file_window`. There's a lot of room for creativity here; you can use `f`
    as a logging handler for a log display. Want a more interactive display?
    Pass `dpg.add_input_text` as the value to the *text_factory* keyword argument
    when initializing the stream.
    """

    def __init__(
        self,
        default_value: Sequence[str] | str = '',
        max_len      : int | None          = None,
        *,
        add_item_callback : Callable[[Self, _T], Any] | None       = None,
        del_item_callback : Callable[[Self, _T], Any] | None       = None,
        set_value_callback: Callable[[Self, _T, Any], Any] | None  = None,
        item_factory      : ItemCommand[_P]                        = dearpygui.add_string_value,
        **kwargs
    ):
        """Args:
            * default_value: Starting value for the item.

            * max_len: The total number of lines the buffer will store. Once at
            capacity, lines are deleted in FIFO order. If unspecified or is None
            (default), the buffer can grow to an arbitrary length.

            * item_factory: Callable that returns a DearPyGui item identifier.

            * add_item_callback: Callable that accepts this item and the new child value
            item as positional arguments. It will be called immediately following the
            value item's creation.

            * del_item_callback: Callable that accepts this item and the child value item
            scheduled for destruction as item as positional arguments. It will be called
            before destroying the value item.

            * set_value_callback: Callable that accepts this item, the child value item
            who's value will be set, and the value to set, while returning the value to
            set. It will be called before updating the value of a value item.

        """
        super().__init__(**kwargs)
        self._max_len = max_len

        self._add_value_item = Method(
            tools.create_function(
                '_new_value_item',
                ('self', "*args", "**kwargs"),
                (
                    'kwargs["parent"] = self',
                    'vitem = item_factory(*args, **kwargs)',
                    'callback(self, vitem)',
                    'return vitem',
                ),
                Item,
                locals={
                    'item_factory': item_factory,
                    'callback': add_item_callback or _null_item_processing,
                }
            ),
            self,
        )
        self._del_value_item = Method(
            tools.create_function(
                '_del_value_item',
                ('self', 'item: int | str', '/'),
                ('callback(self, item)', 'delete_item(item)'),
                Item,
                locals={
                    'callback': del_item_callback or _null_item_processing,
                    'delete_item': api.Registry.delete_item
                }
            ),
            self,
        )
        self._set_value_item = Method(
            tools.create_function(
                '_set_value_item',
                ('self', 'item: int | str', 'value: str', '/'),
                ('set_value(item, callback(self, item, value))',),
                locals={
                    'set_value': api.Item.set_value,
                    'callback': set_value_callback or _null_value_processing,
                }
            ),
            self,
        )

        self._generate_items = functools.partial(_item_generator, self._add_value_item)

        if default_value:
            self.set_value(default_value)

    def __str__(self):
        return '\n'.join(self.get_value())

    def __iter__(self):
        vitems = self.children(1)
        yield from zip(vitems, ItemAPI.get_values(vitems))

    def __setitem__(self, index: SupportsIndex, value: Any):
        ItemAPI.set_value(self.children(1)[index], value)

    def _add_value_item(self, *args, **kwargs) -> Item: ...

    def _del_value_item(self, item: Item, /) -> None: ...

    def _set_value_item(self, item: Item, value: str, /) -> None: ...

    def _generate_items(self, *args, **kwargs) -> Generator[Item, None, None]: ...

    def _truncate_lines(self):
        if self._max_len is not None:
            for vitem in self.children(1)[self._max_len:]:
                self._del_value_item(vitem)

    @property
    def max_len(self) -> int | None:
        return self._max_len
    @max_len.setter
    def max_len(self, value: int | None):
        if value is not None:
            value = int(value)
            if value < 0:
                raise ValueError("`max_len` cannot be less than zero.")
        self._max_len = value

    @override
    def get_value(self) -> list[str]:
        return ItemAPI.get_values(self.children(1))

    @override
    def set_value(self, value: Sequence[Any]):
        vitems = self.children(1)
        for v, item in zip(value, itertools.chain(vitems, self._generate_items())):
            self._set_value_item(item, v)
        # delete leftover children if any
        for item in vitems[len(value):]:
            self._del_value_item(item)
        self._truncate_lines()

    @override
    def delete(self, *, children_only: bool = False, slot: Literal[-1, 0, 1, 2, 3] = -1) -> None:
        self.clear()
        super().delete(children_only=children_only, slot=slot)

    # list-like methods

    @override
    def pop(self, index: SupportsIndex) -> tuple[Item, str]:
        """Remove a line from the buffer at *index*, returning both the value
        item identifier and value."""
        vitem = self.children(1)[index]
        value = ItemAPI.get_value(vitem)
        self._del_value_item(vitem)
        return value

    @override
    def remove(self, value: str) -> None:
        """At the first occurence of *value*, remove the line and value."""
        index = self.get_value().index(value)
        self._del_value_item(self.children(1)[index])

    @override
    def insert(self, index: SupportsIndex, value: Any) -> None:
        """Insert *value* at *index*, offsetting the position of trailing values by 1.
            >>> f = FileStream(["one", "two", "three"])
            >>> f.insert(1, "four")
            >>> arr
            ["one", "four", "two", "three"]
        """
        # Unlike most child items, value items don't have a "before" option.
        # The new item must be added first, then rearrange the children.
        vitems = self.children(1)
        vitem  = self._add_value_item()
        self._set_value_item(vitem, value)
        vitems.insert(index, vitem)
        self.reorder(1, vitems)
        self._truncate_lines()

    @override
    def extend(self, value: Sequence[Any]):
        v = self.get_value()
        v.extend(value)
        self.set_value(v)

    @override
    def append(self, value: Any):
        self._set_value_item(self._add_value_item(), value)
        self._truncate_lines()

    def clear(self, *, use_del_item_callback: bool = False):
        if not use_del_item_callback:
            ItemAPI.delete(self, children_only=True, slot=1)
        else:
            self.set_value('')

    # misc

    def strip(self):
        """Remove all empty lines from the start and end of the buffer.
            >>> f = FileStream(["", "", "foo", "bar", ""])
            >>> f.strip()
            >>> arr
            ["foo", "bar"]
        """
        for item, value in self:
            if value:
                break
            self._del_value_item(item)
        for item, value in self[::-1]:
            if value:
                break
            self._del_value_item(item)

    def write(self, s: str):
        self.append(s)




class ConsoleWindow(items.mvChildWindow):
    """A child window for displaying text logs. Writing to the item updates
    the display with the written text.

    The window does not push to and pop itself from the container stack when
    used as a context manager. Instead, any write to stdout and/or stderr
    will be written to the window instead. This behavior can be altered by
    setting the boolean `.ctx_redirect_stdout` and `.ctx_redirect_stderr`
    attributes on the class or instance.

    Note that when this interface is initialized, a resize handler is
    added to the handler registry of the root parent. The handler registry
    will be created if necessary.

    """
    ctx_redirect_stdout: bool = True
    ctx_redirect_stderr: bool = False

    def __init__(
        self,
        max_len: int | None = None,
        *,
        redirect_stdout: bool | None        = None,
        redirect_stderr: bool | None        = None,
        auto_scroll    : bool               = True,
        wrap_text      : bool               = True,
        text_factory   : ItemCommand[_P]    = dearpygui.add_text,
        buffer_factory : type[FileStream] = FileStream,
        **kwargs,
    ):
        """Args:
            * max_len: The total number of records to display before deleting
            older records in FIFO order.

            * redirect_stdout: If True (class-level default), all writes to stdout
            will instead write to this window when it is used as a context manager.
            If None, the `.ctx_redirect_stdout` attribute will not be set at the
            instance-level.

            * redirect_stderr: If True, all writes to stderr will instead write to
            this window when it is used as a context manager. If None, the
            `.ctx_redirect_stderr` attribute will not be set at the instance-level.

            * auto_scroll: If True (default), the scroll position is set to view the newest
            record when it is added.

            * wrap_text: If True, displayed text will be wrapped when the root
            window parent item is resized. Ignored when *text_factory* does produce
            'mvText' items.

            * text_factory: Creates the text items that the window displays. Supports
            callables that create 'mvText' and 'mvInputText' items.

        """
        super().__init__(**kwargs)
        self.wrap_text    = wrap_text
        self.text_factory = text_factory
        self.auto_scroll  = auto_scroll
        self.value_buffer = buffer_factory(
            max_len=max_len,
            add_item_callback=self._cb_add_record,
            del_item_callback=self._cb_del_record,
            set_value_callback=self._cb_set_record,
        )
        self.__ctx_lock    = threading.RLock()
        self.__ctx_counter = 0
        self.__ctx_stdout  = contextlib.redirect_stdout(self)  # type: ignore
        self.__ctx_stderr  = contextlib.redirect_stderr(self)  # type: ignore
        if redirect_stdout is not None:
            self.ctx_redirect_stdout = redirect_stdout
        if redirect_stderr is not None:
            self.ctx_redirect_stderr = redirect_stderr
        self._attach_resize_handler()


    # ResizeHandler management

    _resize_handler_uuid = 0

    def _destroy_resize_handler(self):
        try:
            ItemAPI.delete(self._resize_handler_uuid)
        except SystemError:
            pass

    def _attach_resize_handler(self):
        # XXX: this needs to be called whenever the console is re-parented!
        if self._resize_handler_uuid:
            self._destroy_resize_handler()
        else:
            # this uuid is recycled and should not change for this object
            self._resize_handler_uuid = api.Registry.create_uuid()
        root_parent   = self.root_parent
        root_handlers = ItemAPI.information(root_parent)['handlers']
        if root_handlers is None:
            root_handlers = items.ItemHandlerRegistry(label='')
            ItemAPI.set_handlers(root_parent, root_handlers)
        items.ResizeHandler(
            label='',
            parent=root_handlers,
            callback=self._cb_resize,
            tag=self._resize_handler_uuid,
        )

    # buffer callbacks

    def _cb_add_record(self, b: FileStream, item: Item):
        text_item = self.text_factory(
            parent=self,
            source=item,
        )
        if self.wrap_text and 'wrap' in ItemAPI.configuration(text_item):
            ItemAPI.configure(text_item, wrap=self.content_region_avail[0])   # type: ignore
        # The 'del_item' callback will receive the source value item but
        # not the text item. Keep a reference to the text item in the value
        # item so it can be easily deleted later.
        ItemAPI.configure(item, user_data=text_item)
        if self.auto_scroll:
            self.y_scroll_pos = -1.0

    def _cb_del_record(self, b: FileStream, item: Item):
        ItemAPI.delete(ItemAPI.configuration(item)['user_data'])

    def _cb_set_record(self, b: FileStream, item: Item, value: str)  -> str:
        # XXX: This is unchanged from the default behavior and can be overridden
        return value

    def _cb_resize(self, sender: Item, *args):
        if self.wrap_text:
            width = self.content_region_avail[0]   # type: ignore
            configuraton = ItemAPI.configuration
            for text in self.children(1):
                if 'wrap' in configuraton(text):
                    try:
                        ItemAPI.configure(text, wrap=width)
                    except SystemError:
                        pass

    # context usage

    __ctx_no_redirect = contextlib.nullcontext()
    __stdout_redirect = __ctx_no_redirect
    __stderr_redirect = __ctx_no_redirect

    def __enter__(self) -> Self:
        self.redirect()
        return self

    def __exit__(self, *args):
        self.restore(*args)

    def redirect(self):
        with self.__ctx_lock:
            if self.__ctx_counter > 1:
                self.__ctx_counter += 1
                return
            self.__ctx_counter = 1
            if self.ctx_redirect_stdout:
                self.__stdout_redirect = self.__ctx_stdout
                self.__stdout_redirect.__enter__()
            if self.ctx_redirect_stderr:
                self.__stderr_redirect = self.__ctx_stderr
                self.__stderr_redirect.__enter__()

    def restore(self, *args):
        with self.__ctx_lock:
            if self.__ctx_counter > 1:
                self.__ctx_counter -= 1
                return
            self.__stdout_redirect.__exit__(*args)
            self.__stdout_redirect = self.__ctx_no_redirect
            self.__stderr_redirect.__exit__(*args)
            self.__stderr_redirect = self.__ctx_no_redirect
            self.__ctx_counter = 0

    # the rest

    @property
    def max_len(self):
        return self.value_buffer.max_len
    @max_len.setter
    def max_len(self, value: Any):
        self.value_buffer.max_len = value

    @override
    def move(self, *, parent: Item = 0, before: Item = 0) -> None:
        super().move(parent=parent, before=before)
        if parent:
            self._attach_resize_handler()

    @override
    def unstage(self):
        super().unstage()
        self._attach_resize_handler()

    @override
    def delete(self, *, children_only: bool = False, slot: Literal[-1, 0, 1, 2, 3] = -1) -> None:
        if not children_only:
            self._destroy_resize_handler()
        self.value_buffer.delete(children_only=children_only, slot=slot)
        super().delete(children_only=children_only, slot=slot)

    @override
    def destroy(self):
        try: self._destroy_resize_handler()
        except: pass
        try: self.value_buffer.destroy()
        except: pass
        super().destroy()

    def clear(self, **kwargs):
        self.value_buffer.clear(use_del_item_callback=True)

    def write(self, s: str):
        return self.value_buffer.write(s)



class PythonConsole(ConsoleWindow):
    """A console window for displaying the output of executed Python
    code.

    See the `.push` and `.compile` methods for more information.
    """
    def __init__(
        self,
        locals : dict[str, Any] | None = None,
        max_len: int | None            = None,
        **kwargs
    ):
        super().__init__(
            max_len,
            redirect_stderr=True,
            redirect_stdout=True,
            auto_scroll=True,
            wrap_text=True,
            text_factory=dearpygui.add_text,
            **kwargs
        )
        self._compiler = codeop.CommandCompiler()

        self.locals = locals if locals is not None else {
            "__name__": "__console__",
            "__doc__" : None,
        }

        with items.mvTheme(label='') as self.theme:
            with items.mvThemeComponent(int(items.mvChildWindow), label=''):
                self.console_color = color.ChildBg(25, 25, 25)
                style.WindowPadding(8, 0)

        with items.mvTheme() as self._stdout_theme:
            with items.mvThemeComponent(label=''):
                self.stdout_color = color.Text(255, 255, 255)
                color.FrameBg(0, 0, 0, 0)

        with items.mvTheme() as self._stderr_theme:
            with items.mvThemeComponent(label=''):
                self.stderr_color = color.Text(255, 20, 20)
                color.FrameBg(0, 0, 0, 0)

    def _cb_set_record(self, b: FileStream, item: Item, value: str) -> str:
        # set the theme of the text
        text_item = ItemAPI.configuration(item)['user_data']
        if value.startswith("Traceback (") or '\nSyntaxError:' in value or '\nIndentationError:' in value:
            theme = self._stderr_theme
        else:
            theme = self._stdout_theme
        ItemAPI.set_theme(text_item, theme)
        return value

    @property
    def filename(self):
        return self.locals.get('__name__', '<console>')

    def _write_syntax_err(self):
        """Format and write the most recent syntax error set and
        write it to the console.
        """
        exc_tp, exc_val, tb  = sys.exc_info()
        sys.last_type        = exc_tp
        sys.last_value       = exc_val
        sys.last_traceback   = tb

        # SyntaxErrors don't actually include the traceback in
        # stderr. However, the traceback lines are formatted
        # the same regardless. Remove the indent(s).
        lines = traceback.format_exception_only(exc_tp, exc_val)
        indent = ""
        for char in lines[0]:
            if not char.isspace():
                break
            indent += char
        for idx, ln in enumerate(lines):
            lines[idx] = ln.removeprefix(indent)

        self.value_buffer.write(''.join(lines))

    def _write_traceback(self):
        """Format the most recent exception set and write it to the
        console.
        """
        sys.last_type, sys.last_value, last_tb = ei = sys.exc_info()
        sys.last_traceback = last_tb
        try:
            # use the next traceback because the first is from *our* code
            lines = traceback.format_exception(ei[0], ei[1], last_tb.tb_next)  # type: ignore
            self.value_buffer.write(''.join(lines))
        finally:
            last_tb = ei = None

    def compile(self, s: str, filename: str = '', mode: str = 'single'):
        """Compile a string as source code.

        The return of this method varies based on the outcome of the compilation:
            * CodeType: The code compiled sucessfully (as the returned object).

            * False: The code was compiled successfully but could not be executed
            because it is incomplete.

            * None: `SyntaxError` was thrown while compiling the code.

        All exceptions that occur from compiling the code are displayed in the
        console and are not raised.
        """
        filename = filename or self.filename
        try:
            code = self._compiler(s, filename, mode)
        except (OverflowError, SyntaxError, ValueError):
            self._write_syntax_err()
            return None   # invalid source code
        if code is None:  # incomplete source code
            return False
        return code

    def push(self, c: types.CodeType | str):
        """Compile a string as source code, execute it, and display the output
        in the console.

        The return of this method varies based on the outcome of the compilation:
            * True: The code compiled sucessfully or was not needed, and was
            executed.

            * False: The code was compiled successfully but could not be executed
            because it is incomplete.

            * None: `SyntaxError` was thrown while compiling the code.

        All exceptions that occur from compiling or executing the code are
        displayed in the console and are not raised.

        Accepts a string or `CodeType` object as an argument. This method will
        execute a `CodeType` object immediately and will always return True.
        """
        if isinstance(c, types.CodeType):
            code = c
        else:
            code = self.compile(c, self.filename)
            if not code:
                return code
        with self:
            try:
                exec(code, self.locals)
            except:
                self._write_traceback()
        return True




@interface.auto_parent(functools.partial(  # type: ignore
    dearpygui.add_window,
    width=600,
    height=400,
    no_scrollbar=True
))
class InteractivePython(items.mvChildWindow):
    """Execute Python code from an emulated interactive session.

    This item will create a root parent window for itself if a parent is
    unspecified or unavailable.

    When this interface is initialized, a resize handler is added to the
    handler registry of the root parent. The handler registry will be
    created if necessary.
    """

    _PROMPT_1 = ">>>"
    _PROMPT_2 = "..."

    def __init__(
        self,
        locals: dict[str, Any] | None = None,
        echo  : bool                  = True,
        **kwargs
    ):
        """Args:
            * locals: Namespace used for lookups.

            * echo: If True, text from the input will be displayed in the
            console when the 'enter' key is pressed.
        """
        if not 'label' in kwargs:
            kwargs['label'] = f'[{type(self).__qualname__}]'
        super().__init__(**kwargs)
        self._input_pending = []

        self.echo = echo

        with items.mvTheme() as self.theme:
            with items.mvThemeComponent():
                style.WindowPadding(0, 0)
                style.FramePadding(4, 4)
                style.ItemSpacing(8, 0)
                color.FrameBg(0, 0, 0, 0)

        with items.mvTheme() as self._echo_theme:
            with items.mvThemeComponent():
                self.echo_color = color.Text(50, 210, 230, 150)  # cyan

        with self:
            # The initial height needs to be very small so the input parent can
            # have room to stretch its' legs for the initial render. Otherwise,
            # the console will squish it and its y_scroll_max will be 0.
            with PythonConsole(locals=locals, height=1) as self.console:
                ...
            with items.mvChildWindow(no_scrollbar=True, border=False, height=24) as self._input_parent:
                self._input_parent.theme = self.theme
                with items.mvGroup(horizontal=True, horizontal_spacing=0):
                    self._input_prompt = items.mvText(self._PROMPT_1, indent=8)  # indent == WindowPadding[0]
                    self.input = items.mvInputText(
                        on_enter=True,
                        callback=self._cb_enter,
                        hint="(enter a partial or complete Python expression/statement)"
                    )

        self._attach_resize_handler()

        self.console.write("Python {} on {}\n{}\n".format(
            sys.version,
            sys.platform,
            'Type "help", "copyright", "credits" or "license" for more information.\n',
        ))

    def _cb_enter(self, sender: Item, app_data: str, *args):
        self._input_pending.append(app_data)
        self.input.value = ''
        if self.echo and app_data:
            self.console.write(f"{self._input_prompt.value} {app_data}")
            ItemAPI.set_theme(self.console.children(1)[-1], self._echo_theme)
        code = self.console.compile('\n'.join(self._input_pending))
        if code:
            self.console.push(code)
            self._input_prompt.value = self._PROMPT_1
            self._input_pending.clear()
        elif code is None:  # SyntaxError, etc.
            self._input_prompt.value = self._PROMPT_1
            self._input_pending.clear()
        else:  # pending
            self._input_prompt.value = self._PROMPT_2
        self.input.focus()

    def _cb_resize(self):
        # XXX: ALL of the input window must be "visible" for the first
        # render so its' y_scroll_max gets set before `_cb_resize` reads
        # it. If it's only viewable off-screen (in the parent's scroll
        # area, etc) y_scroll_max will be initially set to 0. This can
        # be worked around by setting the initial height of the console
        # to *smol-er* to keep it from being a fatass.
        text_wt, text_ht = self.text_size(
            self._PROMPT_1,
            font=self.font or api.Application.get_font()  # type: ignore
        )
        wt_avail, ht_avail = self.content_region_avail  # type: ignore
        # Resize the input so that it's just tall enough for the
        # rendered font. It could be larger or smaller than before;
        # set the height to *smol* and wait a frame (at least), then
        # set the height based off the upd. max scroll position.
        self._input_parent.height = 5
        dearpygui.split_frame(delay=16)
        input_ht = int(self._input_parent.y_scroll_max) + 8   # account for `FramePadding[1]`
        self._input_parent.height = input_ht

        self.input.configure(width=wt_avail - text_wt - 8)  # account for `WindowPadding[0]`

        self.console.height = ht_avail - input_ht

    _resize_handler_uuid = 0

    def _destroy_resize_handler(self):
        try:
            ItemAPI.delete(self._resize_handler_uuid)
        except SystemError:
            pass

    def _attach_resize_handler(self):
        # XXX: this needs to be called whenever the console is re-parented!
        if self._resize_handler_uuid:
            self._destroy_resize_handler()
        else:
            # this uuid is recycled and should not change for this object
            self._resize_handler_uuid = api.Registry.create_uuid()
        root_parent   = self.root_parent
        root_handlers = ItemAPI.information(root_parent)['handlers']
        if root_handlers is None:
            root_handlers = items.ItemHandlerRegistry(label='')
            ItemAPI.set_handlers(root_parent, root_handlers)
        items.ResizeHandler(
            label='',
            parent=root_handlers,
            callback=self._cb_resize,
            tag=self._resize_handler_uuid,
        )

    @override
    def move(self, *, parent: Item = 0, before: Item = 0) -> None:
        super().move(parent=parent, before=before)
        if parent:
            self._attach_resize_handler()

    @override
    def unstage(self):
        super().unstage()
        self._attach_resize_handler()

    @override
    def delete(self, *, children_only: bool = False, slot: Literal[-1, 0, 1, 2, 3] = -1) -> None:
        if not children_only:
            self._destroy_resize_handler()
        self.console.delete(children_only=children_only, slot=slot)
        super().delete(children_only=children_only, slot=slot)

    @override
    def destroy(self):
        try: self._destroy_resize_handler()
        except: pass
        try: self.console.destroy()
        except: pass
        super().destroy()

    def clear_input(self):
        """Empty the input buffer."""
        self._input_pending.clear()





import sys
import abc
import datetime
from numbers import Number
from types import (
    CodeType,
    CodeType as Code,
    EllipsisType,
    GenericAlias,
    MappingProxyType,
    MappingProxyType as MappingProxy,
    MethodType,
    MethodType as Method,
)
from collections.abc import (
    Callable,
    Collection,
    Generator,
    Iterable,
    Iterator,
    Sequence,
    MutableSequence,
    Mapping,
    MutableMapping,
    KeysView,
    ValuesView,
    ItemsView,
)
from typing import (
    Any,
    Required,
    NotRequired,

    SupportsIndex,
    SupportsAbs,
    SupportsBytes,
    SupportsComplex,
    SupportsFloat,
    SupportsInt,
    SupportsRound,

    Self,
    Literal,
    Generic,
    Protocol,
    TypedDict,

    Annotated,
    Concatenate,
    ClassVar,
    TypeAlias,
    TypeVar,
    TypeVarTuple,
    ParamSpec,
    Unpack,
    Union,

    NamedTuple,

    cast,
    final,
    get_overloads,
    overload,
    runtime_checkable,

    TYPE_CHECKING,
)
if sys.version_info < (3, 12):
    try:
        from typing_extensions import override
    except ImportError:
        raise ImportError("Python < 3.12 requires `typing-extensions` module.") from None
else:
    from typing import override
from . import tools

if TYPE_CHECKING:
    from .interface import AppItemType


# Super-primitive types and protocols are declared here.
# Parts of the lib rely on runtime type annotations, so
# `from __future__ import annotations` is rather destru-
# ctive.


_T     = TypeVar("_T")
_N     = TypeVar("_N", bound=Number)
_T_co  = TypeVar("_T_co", covariant=True)
_P     = ParamSpec("_P")




# [ GENERAL ]

Shape = TypeVarTuple("Shape")


@runtime_checkable
class Array(Protocol[*Shape]):  # type: ignore
    __slots__ = ()

    def __iter__(self) -> Iterator[Any]: ...
    @overload
    def __getitem__(self, key: SupportsIndex, /) -> Union[*Shape] | Any: ...
    @overload
    def __getitem__(self, key: slice, /) -> 'Array[*Shape]': ...
    def __len__(self) -> int: ...
    def __contains__(self, value: Any, /) -> bool: ...
    def __reversed__(self) -> Any: ...
    def index(self, value: Any, start: SupportsIndex = 0, stop: SupportsIndex = sys.maxsize, /) -> int: ...
    def count(self, value: Any, /) -> int: ...


@runtime_checkable
class SupportsComp(Protocol):
    __slots__ = ()

    def __lt__(self, other: Any) -> bool: ...
    def __le__(self, other: Any) -> bool: ...
    def __eq__(self, other: Any) -> bool: ...
    def __ge__(self, other: Any) -> bool: ...
    def __gt__(self, other: Any) -> bool: ...


class Property(Protocol[_T_co]):
    def __get__(self, instance: Any, cls: type[Any] | None = None) -> _T_co: ...
    def __set__(self, instance: Any, value: Any): ...


class ValueRange(tuple[int, int]):
    __SENTINEL = object()

    @overload
    def __new__(cls, sequence: Sequence[_N], /) -> Self: ...
    @overload
    def __new__(cls, vmin: _N | Sequence[_N], vmax: _N | Sequence[_N], /) -> Self: ...
    def __new__(cls, vmin: Any, vmax: Any = __SENTINEL, /) -> Self:
        if vmax is cls.__SENTINEL:
            # get both min & max from `vmin`
            try:
                values = min(vmin), max(vmin)
            except TypeError:
                raise TypeError('missing required positional argument `vmax.')
        else:
            values = min(vmin), max(vmax)

        assert values[0] <= values[1]
        return super().__new__(cls, values)

    def __lt__(self, other: SupportsComp | Sequence[int | float]):
        if isinstance(other, Sequence):
            return self[1] < min(other)
        return self[1] < other

    def __le__(self, other: SupportsComp | Sequence[int | float]):
        if isinstance(other, Sequence):
            return self[1] <= min(other)
        return self[1] <= other

    def __eq__(self, other: SupportsComp | Sequence[int | float]):
        if not isinstance(other, Sequence):
            if self[0] == self[1]:
                return self[0] == other
            return False
        return self[0] == min(other) and self[1] == max(other)

    def __ge__(self, other: SupportsComp | Sequence[int | float]):
        if isinstance(other, Sequence):
            return self[0] >= max(other)
        return self[0] >= other

    def __gt__(self, other: SupportsComp | Sequence[int | float]):
        if isinstance(other, Sequence):
            return self[0] < max(other)
        return self[0] < other

    def __iter__(self):
        yield from range(self[0], self[1])

    def __class_getitem__(cls, item: tuple[_N | EllipsisType, _N | EllipsisType]):
        try:
            vmin, vmax = item
        except (TypeError, ValueError):
            raise TypeError('subscript must contain a minimum and maximum value.') from None

        unknown_min = vmin in cls.__CLASS_GETITEM_COMP
        unknown_max = vmax in cls.__CLASS_GETITEM_COMP
        if unknown_min and unknown_max:
            raise TypeError("subscript must contain at least one known boundry.")
        elif unknown_min and isinstance(vmax, SupportsComp):
            tp = type(vmax)
        elif unknown_max and isinstance(vmin, SupportsComp):
            tp = type(vmin)
        else:
            raise TypeError(
                "subscript must contain literal values supporting "
                "rich comparisons or ellipsis (got {vmin!r}, {vmax!r})."
            )
        return Annotated[tp, (vmin, vmax)]

    __CLASS_GETITEM_COMP = (
        ...,
        type(...),
        "...",
        f"{type(...).__name__}"
    )


Color = Array[int, int, int, int | None]
Point = Array[int, int]








# [ ITEM STUFF ]

# XXX: "uuid", "alias", and "item" are thrown around a lot in
# DPG. They are given specific meaning in the context of this
# project:
#   * uuid : A compatible integer id for an item, in-use or
#   otherwise.
#   * alias: A compatible string id for an item, in-use or
#   otherwise.
#   * item : A uuid or alias tied to an active widget in the
#   registry.
ItemAlias  = str
ItemUUID   = int
Item       = ItemUUID | ItemAlias

Interface = TypeVar("Interface", bound='AppItemType')
ItemType  = TypeVar("ItemType", bound=type['AppItemType'])


class ItemCommand(Protocol[_P]):
    __name__: str
    def __call__(self, *, tag: Item = 0, **kwargs) -> Item: ...





class _ItemCallback0(Protocol):
    __code__: CodeType
    def __call__(self, /) -> Any: ...

class _ItemCallback1(Protocol):
    __code__: CodeType
    def __call__(self, __sender: Item, /) -> Any: ...

class _ItemCallback2(Protocol):
    __code__: CodeType
    def __call__(self, __sender: Item, __app_data: Any, /) -> Any: ...

class _ItemCallback3(Protocol):
    __code__: CodeType
    def __call__(self, __sender: Item, __app_data: Any, __user_data: Any, /) -> Any: ...

class _ItemCallback4(Protocol):
    __code__: CodeType
    def __call__(self, /, *args: Any) -> Any: ...

ItemCallback = TypeVar(
    "ItemCallback",
    _ItemCallback0,
    _ItemCallback1,
    _ItemCallback2,
    _ItemCallback3,
    _ItemCallback4,
)




class _ItemProperty(Generic[_T]):
    __slots__ = ("_key",)

    def __init__(self, key: str = '', /):
        self._key = key

    def __set_name__(self, cls: type['ItemInterface'], name: str):
        self._key = self._key or name

    def __set__(self, instance: 'ItemInterface', value: Any):
        raise AttributeError(
            f'{type(instance).__qualname__!r} object attribute {self._key!r} is read-only.'
        )


class ItemConfig(_ItemProperty[_T]):
    """Item interface data descriptor. Uses the interface's `.configuration`
    and `.configure` hooks.
    """
    __slots__ = ('__set__',)

    __set__: Callable[..., None]

    def __set_name__(self, cls: type['ItemInterface'], name: str):
        self._key = self._key or name
        # Dynamically generated for performance. 5-million calls
        # ~ 1 second w/potato CPU -- it doesn't get any faster.
        # XXX: Unfortunately, `__get__` can't have the same treatment.
        #      Python will find the `__get__` member descriptor and,
        #      since it's a dunder method, will call the class-level
        #      member i.e. `member_descriptor.__call__`, which doesn't
        #      exist. Any other dunder would work fine, just not `__get__`.
        __set__ = tools.create_function(
            '__set__',
            ('self', 'instance', 'value'),
            (f"instance.configure({self._key}=value)",)
        )
        self.__set__ = MethodType(__set__, self)

    def __get__(self, instance: 'ItemInterface', cls: type['ItemInterface'] | None = None) -> _T:
        if instance is None:
            return self  # type: ignore
        return instance.configuration()[self._key]



class ItemInfo(_ItemProperty[_T]):
    """Item interface descriptor. Uses the interface's `.information` hook.
    """
    __slots__ = ()

    def __get__(self, instance: 'ItemInterface', cls: type['ItemInterface'] | None = None) -> _T:
        if instance is None:
            return self  # type: ignore
        return instance.information()[self._key]


class ItemState(_ItemProperty[_T]):
    """Item interface descriptor. Uses the interface's `.state` hook.
    """
    __slots__ = ()

    def __get__(self, instance: 'ItemInterface', cls: type['ItemInterface'] | None = None) -> _T | None:
        if instance is None:
            return self  # type: ignore
        return instance.state().get(self._key, None)




null_itemtype = 0, "mvAppItemType::mvAppItem"  # 0 == dearpygui.mvAll

def null_command(*args, tag: Item = 0, **kwargs) -> Item:
    """Default 'command' value for item interface types."""
    return tag


class ItemInterfaceMeta(type(Protocol), abc.ABCMeta):
    __slots__ = ()

    identity: tuple[int, str]
    command : ItemCommand

    def __new__(__mcls: type[Self], __name: str, __bases: tuple[type, ...], __namespace: dict[str, Any], **kwargs: Any) -> Self:
        cls = super().__new__(__mcls, __name, __bases, __namespace, **kwargs)
        if not hasattr(cls, 'command'):
            setattr(cls, 'command', staticmethod(null_command))
        if not hasattr(cls, 'identity'):
            setattr(cls, 'identity', (-hash(cls), cls.__qualname__))
        return cls

    def __int__(self) -> int:
        return self.identity[0]

    def __str__(self) -> str:
        return self.identity[1]

    def __repr__(self) -> str:
        return f"<class {self.__qualname__!r}>"

    def register(self, cls: type[_T]) -> type[_T]:
        # This should only be used internally, and only a couple of
        # times (max).
        assert isinstance(cls, ItemInterface)
        return super().register(cls)


@runtime_checkable
class ItemInterface(Protocol, metaclass=ItemInterfaceMeta):
    """Minimal item API for non-item types."""
    __slots__ = ()

    def __repr__(self):
        config = ', '.join(f"{k}={str(v)!r}" for k, v in self.configuration().items())
        return f"{type(self).__qualname__}({config})"

    @property
    def tag(self) -> ItemUUID:
        return -hash(self)  # signed -- avoids DPG uuid conflicts

    @property
    def alias(self) -> ItemAlias:
        return str(self.tag)

    parent: Item | None = None

    @abc.abstractmethod
    def configure(self, **kwargs) -> None: ...

    @abc.abstractmethod
    def configuration(self) -> Mapping[str, Any]: ...

    def information(self) -> Mapping[str, Any]:
        info = dict(ITEM_INFO_TEMPLATE)
        info['type'] = self.identity[1]  # type:ignore
        return info

    def state(self) -> Mapping[str, Any]:
        return dict(ITEM_STATE_TEMPLATE)

# Only virtual subclasses (`AppItemType`) need to define these, as
# the metaclass ensures that subclasses have these attributes. I
# don't want pyright complaining about it all the friggin' time.
ItemInterface.__annotations__.update(
    identity=tuple[int, str],
    command=ItemCommand,
)



class ItemInfoDict(TypedDict):
    children                        : Mapping[Literal[0, 1, 2, 3], list[Item]]
    type                            : str
    target                          : int
    parent                          : Item | None
    theme                           : Item | None
    handlers                        : Item | None
    font                            : Item | None
    container                       : bool
    hover_handler_applicable        : bool
    active_handler_applicable       : bool
    focus_handler_applicable        : bool
    clicked_handler_applicable      : bool
    visible_handler_applicable      : bool
    edited_handler_applicable       : bool
    activated_handler_applicable    : bool
    deactivated_handler_applicable  : bool
    deactivatedae_handler_applicable: bool
    toggled_open_handler_applicable : bool
    resized_handler_applicable      : bool

ITEM_INFO_TEMPLATE: ItemInfoDict = MappingProxyType(
    ItemInfoDict(
        children=MappingProxyType(dict.fromkeys(range(4), ())),  # type:ignore
        type="",
        target=1,
        parent=None,
        theme=None,
        handlers=None,
        font=None,
        container=False,
        hover_handler_applicable=False,
        active_handler_applicable=False,
        focus_handler_applicable=False,
        clicked_handler_applicable=False,
        visible_handler_applicable=False,
        edited_handler_applicable=False,
        activated_handler_applicable=False,
        deactivated_handler_applicable=False,
        deactivatedae_handler_applicable=False,
        toggled_open_handler_applicable=False,
        resized_handler_applicable=False,
    )
)


class ItemStateDict(TypedDict):
    ok                   : Required[bool]
    pos                  : Array[int, int] | None
    hovered              : bool | None
    active               : bool | None
    focused              : bool | None
    clicked              : bool | None
    left_clicked         : bool | None
    right_clicked        : bool | None
    middle_clicked       : bool | None
    visible              : bool | None
    edited               : bool | None
    activated            : bool | None
    deactivated          : bool | None
    resized              : bool | None
    rect_min             : Array[int, int] | None
    rect_max             : Array[int, int] | None
    rect_size            : Array[int, int] | None
    content_region_avail : Array[int, int] | None

ITEM_STATE_TEMPLATE: ItemStateDict = MappingProxyType(  # type:ignore
    ItemStateDict(
        ok=True,
        pos=None,
        hovered=None,
        active=None,
        focused=None,
        clicked=None,
        left_clicked=None,
        right_clicked=None,
        middle_clicked=None,
        visible=None,
        edited=None,
        activated=None,
        deactivated=None,
        resized=None,
        rect_min=None,
        rect_max=None,
        rect_size=None,
        content_region_avail=None,
    )
)






# Dear PyGui buffer types

class mvBuffer(Protocol):
    """Protocol for Dear PyGui's `mvBuffer` type."""
    def __init__(self, length: int = 0) -> None: ...
    def __str__(self) -> str: ...
    def __len__(self) -> int: ...
    def __getitem__(self, __i: SupportsIndex, /) -> float: ...
    def __setitem__(self, __i: SupportsIndex, __value: float, /) -> None: ...
    def get_width(self) -> int: ...
    def get_height(self) -> int: ...
    def clear_value(self, __initial_value: float, /) -> None: ...


class mvVec4(Protocol):
    """Protocol for Dear PyGui's 4-dimensional vector type `mvVec4`."""
    def __init__(self, x: float = 0.0, y: float = 0.0, z: float = 0.0, w: float = 0.0) -> None: ...
    def __str__(self) -> str: ...
    def __len__(self) -> Literal[4]: ...
    def __getitem__(self, __i: SupportsIndex, /) -> float: ...
    def __setitem__(self, __i: SupportsIndex, __value: float, /) -> None: ...
    def __add__(self, __other: Self, /) -> Self: ...
    def __radd__(self, __other: Self, /) -> Self: ...
    def __sub__(self, __other: Self, /) -> Self: ...
    def __rsub__(self, __other: Self, /) -> Self: ...
    def __mul__(self, __other: Self | float, /) -> Self: ...
    def __rmul__(self, __other: Self | float, /) -> Self: ...


class mvMat4(Protocol):
    """Protocol for Dear PyGui's 4x4 matrix type `mvMat4`."""
    def __init__(self, km00: float = 0.0, km01: float = 0.0, km02: float = 0.0, km03: float = 0.0, km10: float = 0.0, km11: float = 0.0, km12: float = 0.0, km13: float = 0.0, km20: float = 0.0, km21: float = 0.0, km22: float = 0.0, km23: float = 0.0, km30: float = 0.0, km31: float = 0.0, km32: float = 0.0, km33: float = 0.0) -> None: ...
    def __str__(self) -> str: ...
    def __len__(self) -> Literal[16]: ...
    def __getitem__(self, __i: SupportsIndex, /) -> float: ...
    def __setitem__(self, __i: SupportsIndex, __value: float, /) -> None: ...
    def __add__(self, __other: Self, /) -> Self: ...
    def __radd__(self, __other: Self, /) -> Self: ...
    def __sub__(self, __other: Self, /) -> Self: ...
    def __rsub__(self, __other: Self, /) -> Self: ...
    def __mul__(self, __other: Self | float, /) -> Self: ...
    def __rmul__(self, __other: Self | float, /) -> Self: ...



import enum
from dearpygui import dearpygui


class Platform(enum.IntEnum):
    WINDOWS =  dearpygui.mvPlatform_Windows   # 0
    MAC     =  dearpygui.mvPlatform_Apple     # 1
    LINUX   =  dearpygui.mvPlatform_Linux     # 2
    OTHER   = -1

    @classmethod
    def _missing_(cls, value: object):
        if isinstance(value, int):
            return cls.OTHER
        return None


class MouseInput(enum.IntEnum):
    ANY        = -1
    LEFT   = L = dearpygui.mvMouseButton_Left    # 0
    RIGHT  = R = dearpygui.mvMouseButton_Right   # 1
    MIDDLE = M = dearpygui.mvMouseButton_Middle  # 2
    X1         = dearpygui.mvMouseButton_X1      # 3
    X2         = dearpygui.mvMouseButton_X2      # 4


class KeyInput(enum.IntEnum):
    ANY                      = -1
    BREAK                    = 3
    BACKSPACE                = 8
    TAB                      = 9
    CLEAR                    = 12
    RETURN = ENTER           = 13
    SHIFT                    = 16
    CTRL                     = 17
    ALT                      = 18
    PAUSE                    = 19
    CAPS_LOCK                = 20
    ESC                      = 27
    SPACEBAR = SPACE         = 32
    PAGE_UP = PG_UP          = 33
    PAGE_DN = PG_DN          = 34
    END                      = 35
    HOME                     = 36
    ARROW_LEFT               = 37
    ARROW_RIGHT              = 38
    ARROW_UP                 = 39
    ARROW_DN                 = 40
    SELECT                   = 41
    PRINT                    = 42
    EXEC                     = 43
    PRINTSCREEN = PRTSCR     = 44
    INSERT = INS             = 45
    DELETE = DEL             = 46
    HELP                     = 47
    DIGIT_0                  = 48
    DIGIT_1                  = 49
    DIGIT_2                  = 50
    DIGIT_3                  = 51
    DIGIT_4                  = 52
    DIGIT_5                  = 53
    DIGIT_6                  = 54
    DIGIT_7                  = 55
    DIGIT_8                  = 56
    DIGIT_9                  = 57
    A                        = 65
    B                        = 66
    C                        = 67
    D                        = 68
    E                        = 69
    F                        = 70
    G                        = 71
    H                        = 72
    I                        = 73
    J                        = 74
    K                        = 75
    L                        = 76
    M                        = 77
    N                        = 78
    O                        = 79
    P                        = 80
    Q                        = 81
    R                        = 82
    S                        = 83
    T                        = 84
    U                        = 85
    V                        = 86
    W                        = 87
    X                        = 88
    Y                        = 89
    Z                        = 90
    L_META = L_WIN           = 91
    R_META = R_WIN           = 92
    APPS                     = 93
    SLEEP                    = 95
    NUMPAD_0                 = 96
    NUMPAD_1                 = 97
    NUMPAD_2                 = 98
    NUMPAD_3                 = 99
    NUMPAD_4                 = 100
    NUMPAD_5                 = 101
    NUMPAD_6                 = 102
    NUMPAD_7                 = 103
    NUMPAD_8                 = 104
    NUMPAD_9                 = 105
    NUMPAD_MUL               = 106
    NUMPAD_ADD               = 107
    NUMPAD_SEP               = 108
    NUMPAD_SUB               = 109
    NUMPAD_DEC               = 110
    NUMPAD_DIV               = 111
    F1                       = 112
    F2                       = 113
    F3                       = 114
    F4                       = 115
    F5                       = 116
    F6                       = 117
    F7                       = 118
    F8                       = 119
    F9                       = 120
    F10                      = 121
    F11                      = 122
    F12                      = 123
    F13                      = 124
    F14                      = 125
    F15                      = 126
    F16                      = 127
    F17                      = 128
    F18                      = 129
    F19                      = 130
    F20                      = 131
    F21                      = 132
    F22                      = 133
    F23                      = 134
    F24                      = 135
    NUM_LOCK                 = 144
    SCR_LOCK                 = 145
    L_SHIFT                  = 160
    R_SHIFT                  = 161
    L_CTRL                   = 162
    R_CTRL                   = 163
    L_MENU                   = 164
    R_MENU                   = 165
    BROWSER_BACK             = 166
    BROWSER_FORWARD          = 167
    BROWSER_REFRESH          = 168
    BROWSER_STOP             = 169
    BROWSER_SEARCH           = 170
    BROWSER_FAVS             = 171
    BROWSER_HOME             = 172
    VOL_MUTE                 = 173
    VOL_DN                   = 174
    VOL_UP                   = 175
    MEDIA_TRACK_NEXT         = 176
    MEDIA_TRACK_PREV         = 177
    MEDIA_STOP               = 178
    MEDIA_PLAY = MEDIA_PAUSE = 179
    LAUNCH_MAIL              = 180
    MEDIA_SELECT             = 181
    LAUNCH_APP1              = 182
    LAUNCH_APP2              = 183
    SEMICOLON                = 186
    PLUS                     = 187
    COMMA                    = 188
    MINUS                    = 189
    PERIOD                   = 190
    FORWARD_SLASH            = 191
    TILDE                    = 192  # ~
    BRACKET_OPEN             = 219  # [
    BACKSLASH                = 220  #
    BRACKET_CLOSE            = 221
    QUOTE                    = 222
    INTL_BACKSLASH           = 226  # \
    UNIDENTIFIED             = 255


class ThemeCategory(enum.IntEnum):
    CORE = dearpygui.mvThemeCat_Core   # 0
    PLOT = dearpygui.mvThemeCat_Plots  # 1
    NODE = dearpygui.mvThemeCat_Nodes  # 2


class NodeLink(enum.IntEnum):
    INPUT  = dearpygui.mvNode_Attr_Input
    OUTPUT = dearpygui.mvNode_Attr_Output
    STATIC = dearpygui.mvNode_Attr_Static


class NodeMiniMapLoc(enum.IntEnum):
    BOTTOM_LEFT  = BL = dearpygui.mvNodeMiniMap_Location_BottomLeft
    BOTTOM_RIGHT = BR = dearpygui.mvNodeMiniMap_Location_BottomRight
    TOP_RIGHT    = TR = dearpygui.mvNodeMiniMap_Location_TopRight
    TOP_LEFT     = TL = dearpygui.mvNodeMiniMap_Location_TopLeft


class PlotAxis(enum.IntEnum):
    X = dearpygui.mvXAxis
    Y = dearpygui.mvYAxis


class PlotLocation(enum.IntEnum):
    CENTER    = C  = dearpygui.mvPlot_Location_Center
    NORTH     = N  = dearpygui.mvPlot_Location_North
    NORTHEAST = NE = dearpygui.mvPlot_Location_NorthEast
    EAST      = E  = dearpygui.mvPlot_Location_East
    SOUTHEAST = SE = dearpygui.mvPlot_Location_SouthEast
    SOUTH     = S  = dearpygui.mvPlot_Location_South
    SOUTHWEST = SW = dearpygui.mvPlot_Location_SouthWest
    WEST      = W  = dearpygui.mvPlot_Location_West
    NORTHWEST = NW = dearpygui.mvPlot_Location_NorthWest


class TablePolicy(enum.IntEnum):
    FIXED_FIT    = dearpygui.mvTable_SizingFixedFit
    FIXED_SAME   = dearpygui.mvTable_SizingFixedSame
    STRETCH_PROP = dearpygui.mvTable_SizingStretchProp
    STRETCH_SAME = dearpygui.mvTable_SizingStretchSame


class ColorFormat(enum.IntEnum):
    RGBA = dearpygui.mvFormat_Float_rgba
    RGB  = dearpygui.mvFormat_Float_rgb


class PlotMarker(enum.IntEnum):
    NONE     = dearpygui.mvPlotMarker_None
    CIRCLE   = dearpygui.mvPlotMarker_Circle
    SQUARE   = dearpygui.mvPlotMarker_Square
    DIAMOND  = dearpygui.mvPlotMarker_Diamond
    UP       = dearpygui.mvPlotMarker_Up
    DOWN     = dearpygui.mvPlotMarker_Down
    LEFT     = dearpygui.mvPlotMarker_Left
    RIGHT    = dearpygui.mvPlotMarker_Right
    CROSS    = dearpygui.mvPlotMarker_Cross
    PLUS     = dearpygui.mvPlotMarker_Plus
    ASTERISK = dearpygui.mvPlotMarker_Asterisk


class PlotBin(enum.IntEnum):
    SQRT    = dearpygui.mvPlotBin_Sqrt
    STURGES = dearpygui.mvPlotBin_Sturges
    RICE    = dearpygui.mvPlotBin_Rice
    SCOTT   = dearpygui.mvPlotBin_Scott


class ColorEdit(enum.IntEnum):
    ALPHA_PREVIEW_NONE = dearpygui.mvColorEdit_AlphaPreviewNone
    ALPHA_PREVIEW      = dearpygui.mvColorEdit_AlphaPreview
    ALPHA_PREVIEW_HALF = dearpygui.mvColorEdit_AlphaPreviewHalf
    UINT8              = dearpygui.mvColorEdit_uint8
    FLOAT              = dearpygui.mvColorEdit_float
    RGB                = dearpygui.mvColorEdit_rgb
    HSV                = dearpygui.mvColorEdit_hsv
    HEX                = dearpygui.mvColorEdit_hex
    INPUT_RGB          = dearpygui.mvColorEdit_input_rgb
    INPUT_HSV          = dearpygui.mvColorEdit_input_hsv


class PlotColorMap(enum.IntEnum):
    DEFAULT  = dearpygui.mvPlotColormap_Default
    DEEP     = dearpygui.mvPlotColormap_Deep
    DARK     = dearpygui.mvPlotColormap_Dark
    PASTEL   = dearpygui.mvPlotColormap_Pastel
    PAIRED   = dearpygui.mvPlotColormap_Paired
    VIRIDIS  = dearpygui.mvPlotColormap_Viridis
    PLASMA   = dearpygui.mvPlotColormap_Plasma
    HOT      = dearpygui.mvPlotColormap_Hot
    COOL     = dearpygui.mvPlotColormap_Cool
    PINK     = dearpygui.mvPlotColormap_Pink
    JET      = dearpygui.mvPlotColormap_Jet
    TWILIGHT = dearpygui.mvPlotColormap_Twilight
    RDBU     = dearpygui.mvPlotColormap_RdBu
    BRBG     = dearpygui.mvPlotColormap_BrBG
    PIYG     = dearpygui.mvPlotColormap_PiYG
    SPECTRAL = dearpygui.mvPlotColormap_Spectral
    GREYS    = dearpygui.mvPlotColormap_Greys


class ColorPickerDisplay(enum.IntEnum):
    BAR   = dearpygui.mvColorPicker_bar
    WHEEL = dearpygui.mvColorPicker_wheel


class TabOrder(enum.IntEnum):
    REORDERABLE = dearpygui.mvTabOrder_Reorderable
    FIXED       = dearpygui.mvTabOrder_Fixed
    LEADING     = dearpygui.mvTabOrder_Leading
    TRAILING    = dearpygui.mvTabOrder_Trailing


class TimeUnit(enum.IntEnum):
    MICROSECOND = uS = dearpygui.mvTimeUnit_Us
    MILLISECOND = mS = dearpygui.mvTimeUnit_Ms
    SECOND      =  S = dearpygui.mvTimeUnit_S
    MINUTE      =  M = dearpygui.mvTimeUnit_Min
    HOUR        = HR = dearpygui.mvTimeUnit_Hr
    DAY              = dearpygui.mvTimeUnit_Day
    MONTH            = dearpygui.mvTimeUnit_Mo
    YEAR             = dearpygui.mvTimeUnit_Yr


class DatePickerLevel(enum.IntEnum):
    DAY   = dearpygui.mvDatePickerLevel_Day
    MONTH = dearpygui.mvDatePickerLevel_Month
    YEAR  = dearpygui.mvDatePickerLevel_Year


class CullMode(enum.IntEnum):
    NONE  = dearpygui.mvCullMode_None
    BACK  = dearpygui.mvCullMode_Back
    FRONT = dearpygui.mvCullMode_Front


class FontRangeHint(enum.IntEnum):
    DEFAULT                   = dearpygui.mvFontRangeHint_Default
    JAPANESE                  = dearpygui.mvFontRangeHint_Japanese
    KOREAN                    = dearpygui.mvFontRangeHint_Korean
    CHINESE_FULL              = dearpygui.mvFontRangeHint_Chinese_Full
    CHINESE_SIMPLIFIED_COMMON = dearpygui.mvFontRangeHint_Chinese_Simplified_Common
    CYRILLIC                  = dearpygui.mvFontRangeHint_Cyrillic
    THAI                      = dearpygui.mvFontRangeHint_Thai
    VIETNAMESE                = dearpygui.mvFontRangeHint_Vietnamese


class NodePinShape(enum.IntEnum):
    CIRCLE          = dearpygui.mvNode_PinShape_Circle
    CIRCLE_FILLED   = dearpygui.mvNode_PinShape_CircleFilled
    TRIANGLE        = dearpygui.mvNode_PinShape_Triangle
    TRIANGLE_FILLED = dearpygui.mvNode_PinShape_TriangleFilled
    QUAD            = dearpygui.mvNode_PinShape_Quad
    QUAD_FILLED     = dearpygui.mvNode_PinShape_QuadFilled


class Direction(enum.IntEnum):
    NONE  = dearpygui.mvDir_None
    LEFT  = dearpygui.mvDir_Left
    RIGHT = dearpygui.mvDir_Right
    UP    = dearpygui.mvDir_Up
    DOWN  = dearpygui.mvDir_Down


class ComboHeight(enum.IntEnum):
    SMALL   = dearpygui.mvComboHeight_Small
    REGULAR = dearpygui.mvComboHeight_Regular
    LARGE   = dearpygui.mvComboHeight_Large
    LARGEST = dearpygui.mvComboHeight_Largest


class GraphicsBackend(enum.IntEnum):
    D3D11 = DX11 = dearpygui.mvGraphicsBackend_D3D11
    D3D12 = DX12 = dearpygui.mvGraphicsBackend_D3D12
    VULKAN       = dearpygui.mvGraphicsBackend_VULKAN
    METAL        = dearpygui.mvGraphicsBackend_METAL
    OPENGL       = dearpygui.mvGraphicsBackend_OPENGL



"""Grid-layout manager for Dear PyPixl and Dear PyGui."""
import types
from dearpygui._dearpygui import (  # using the "raw" API to help w/performance in some places
    get_item_state as _get_item_state,
    get_item_configuration as _get_item_size,
    get_viewport_configuration as _get_viewport_size,
    configure_item as _set_item_size,
)
from ._dearpypixl.common import (
    Item,

    Any,
    Array,
    Self,
    NamedTuple,
    Protocol,
    Callable,
    Literal,
    Mapping,
    Generic,
    Iterator,
    Collection,
    overload,

    TypeVar,
)
from ._dearpypixl.api import Registry as RegistryAPI
from ._dearpypixl import errors


# TODO: add thread lock


Rect = Array[float, float, float, float]  # width, height, x_pos, y_pos




####################################################
########### ITEMSIZE GET/SET FUNCTIONS #############
####################################################

class FGetSize(Protocol):
    """A function that returns the size of an item."""
    def __call__(self, item: Item) -> tuple[int, int]: ...

def size_from_size(item: Item) -> tuple[int, int]:
    cfg = _get_item_size(item)
    return cfg["width"], cfg["height"]

def size_from_rect(item: Item) -> tuple[int, int]:
    return _get_item_state(item)["rect_size"]

def size_from_viewport(item: Any = None) -> tuple[int, int]:
    cfg = _get_viewport_size("DPG NOT USED YET")
    return cfg['width'], cfg['height']


class FSetSize(Protocol):
    """A function that sets the size and position of an item."""
    def __call__(self, item: Item, *, width: int, height: int, pos: Array[int, int]) -> Any: ...




####################################################
###### ANCHORS & ANCHOR SIZE/POS CALCULATORS #######
####################################################

_ANCHOR_MAP = {}


class FAnchor(Protocol):
    def __call__(self, item_wt: float, item_ht: float, cell_x: float, cell_y: float, cell_wt: float, cell_ht: float, /) -> tuple[int, int]: ...


def _register_anchors(*anchor: str) -> Callable[[FAnchor], FAnchor]:
    def register_anchor_fn(fn: FAnchor) -> FAnchor:
        for s in anchor:
            _ANCHOR_MAP[s.lower()] = fn  # type: ignore
        return fn
    return register_anchor_fn


@_register_anchors("n", "north")
def _calc_anchor_N(item_wt: float, item_ht: float, cell_x: float, cell_y: float, cell_wt: float, cell_ht: float) -> tuple[int, int]:
    return int((cell_wt - item_wt) / 2 + cell_x), int(cell_y)  # center x

@_register_anchors("ne", "northeast")
def _calc_anchor_NE(item_wt: float, item_ht: float, cell_x: float, cell_y: float, cell_wt: float, cell_ht: float) -> tuple[int, int]:
    return int((cell_wt - item_wt) + cell_x), int(cell_y)

@_register_anchors("e", "east")
def _calc_anchor_E(item_wt: float, item_ht: float, cell_x: float, cell_y: float, cell_wt: float, cell_ht: float) -> tuple[int, int]:
    return int((cell_wt - item_wt) + cell_x), int((cell_ht - item_ht) / 2 + cell_y)  # center y

@_register_anchors("se", "southeast")
def _calc_anchor_SE(item_wt: float, item_ht: float, cell_x: float, cell_y: float, cell_wt: float, cell_ht: float) -> tuple[int, int]:
    return int((cell_wt - item_wt)     + cell_x), int((cell_ht - item_ht)     + cell_y)

@_register_anchors("s", "south")
def _calc_anchor_S(item_wt: float, item_ht: float, cell_x: float, cell_y: float, cell_wt: float, cell_ht: float) -> tuple[int, int]:
    return int((cell_wt - item_wt) / 2 + cell_x), int((cell_ht - item_ht)     + cell_y)  # center x

@_register_anchors("sw", "southwest")
def _calc_anchor_SW(item_wt: float, item_ht: float, cell_x: float, cell_y: float, cell_wt: float, cell_ht: float) -> tuple[int, int]:
    return int(cell_x), int((cell_ht - item_ht) + cell_y)

@_register_anchors("w", "west")
def _calc_anchor_W(item_wt: float, item_ht: float, cell_x: float, cell_y: float, cell_wt: float, cell_ht: float) -> tuple[int, int]:
    return int(cell_x), int((cell_ht - item_ht) / 2 + cell_y)  # center y

@_register_anchors("nw", "northwest")
def _calc_anchor_NW(item_wt: float, item_ht: float, cell_x: float, cell_y: float, cell_wt: float, cell_ht: float) -> tuple[int, int]:
    return int(cell_x), int(cell_y)

@_register_anchors("c", "center")
def _calc_anchor_C(item_wt: float, item_ht: float, cell_x: float, cell_y: float, cell_wt: float, cell_ht: float) -> tuple[int, int]:
    return int((cell_wt - item_wt) / 2 + cell_x), int((cell_ht - item_ht) / 2 + cell_y)  # center x & y



_ANCHOR_MAP: Mapping[str, FAnchor] = types.MappingProxyType(_ANCHOR_MAP)




####################################################
########### MISC TYPES/HELPER FUNCTIONS ############
####################################################

# These are decoupled from `Grid` to make it easier to monkeypatch
# in more performant alternatives later.

def normalize_cellspan(r1: int, c1: int, r2: int, c2: int, rows: int, cols: int):
    """Corrects backwards/reversed start-stop slot ranges and returns the positive
    indexes of a cellspan.

    Args:
        * r1: A positive or negative integer representing the index of the starting
        row of the cellspan. Its' absolute value should be within `range(rows)`.

        * c1: A positive or negative integer representing the index of the starting
        column of the cellspan. Its' absolute value should be within `range(cols)`.

        * r2: A positive or negative integer representing the index of the ending
        row of the cellspan. Its' absolute value should be within `range(rows)`.

        * c2: A positive or negative integer representing the index of the ending
        column of the cellspan. Its' absolute value should be within `range(cols)`.

        * rows: The total number of rows.

        * cols: The total number of columns.
    """
    r1 %= rows
    c1 %= cols
    r2 %= rows
    c2 %= cols
    if r1 > r2:
        r1, r2 = r2, r1
    if c1 > c2:
        c1, c2 = c2, c1
    return r1, c1, r2, c2


def draw_cells(
    rect_size: tuple[int, int],
    rect_pad : tuple[float, float, float, float],
    cell_pad : tuple[float, float, float, float],
    x_axis   : 'GridAxis',
    y_axis   : 'GridAxis',
) -> dict[tuple[int, int], tuple[float, ...]]:
    """Args:
        * rect_size: Width and height of the grid's parent/target.

        * rect_pad: The amount of whitespace (in pixels) between the grid's
        parent/target and the outer-most cells, as `(left, right, upper, lower)`.

        * cell_pad: The amount of whitespace (in pixels) between cells, as
        `(left, right, upper, lower)`.

        * x_axis: `GridAxis` object containing horizontal slot settings.

        * y_axis: `GridAxis` object containing vertical slot settings.
    """
    rect_x1_pad = rect_pad[0]
    rect_y1_pad = rect_pad[2]
    rect_wt     = rect_size[1] - rect_y1_pad - rect_pad[3]
    rect_ht     = rect_size[0] - rect_x1_pad - rect_pad[1]

    weight_wt = max(0, (rect_ht - y_axis.min_size)) / max(1, y_axis.weight)  # `max(1, ...)` prevents `ZeroDivisionError`
    weight_ht = max(0, (rect_wt - x_axis.min_size)) / max(1, x_axis.weight)

    # The size of each cell is calculated using half of each padding
    # value; two cells side-by-side have the "full" amount of padding
    # between them.
    cell_x1_pad = cell_pad[0] / 2
    cell_y1_pad = cell_pad[2] / 2
    cell_x_pad  = cell_x1_pad + (cell_pad[1] / 2.0)
    cell_y_pad  = cell_y1_pad + (cell_pad[3] / 2.0)

    cell_map = {}
    alloc_ht = 0.0

    rows = x_axis._members
    cols = [*enumerate(y_axis._members)]  # iterated over many times, unlike `rows`
    for row, row_cfg in enumerate(rows):
        # calculate each cell's vertical rect
        row_ht     = row_cfg.size or weight_ht * row_cfg.weight  # FIXED or SIZED policy
        cell_y_pos = rect_y1_pad + alloc_ht + cell_y1_pad
        cell_ht    = row_ht - cell_y_pad
        alloc_wt   = 0.0
        for col, col_cfg in cols:
            # calculate each cell's horizontal rect
            col_wt     = col_cfg.size or weight_wt * col_cfg.weight  # FIXED or SIZED policy
            cell_x_pos = rect_x1_pad + alloc_wt + cell_x1_pad
            cell_wt   = col_wt - cell_x_pad
            alloc_wt += col_wt
            cell_map[(row, col)] = (
                cell_x_pos,
                cell_y_pos,
                cell_wt,
                cell_ht,
            )
        alloc_ht += row_ht
    return cell_map


def redraw_grid(grid: 'Grid', *args):
    """Update the position and size of items managed by the grid."""
    cells = draw_cells(
        grid._sizer(grid._parent),
        grid._padding,
        grid._spacing,
        grid._rows,
        grid._cols,
    )
    row_len = len(grid._rows)
    col_len = len(grid._cols)
    del_items = None
    for item, coords1, coords2, item_width, item_height, anchor, size_setter, in grid._members.values():
        r1, c1, r2, c2 = normalize_cellspan(*coords1, *coords2, row_len, col_len)
        x_pos , y_pos , width1, height1 = cells[(r1, c1)]
        x_offs, y_offs, width2, height2 = cells[(r2, c2)]
        # fix offset when drawing over a cell range
        cell_width  = x_offs + width2  - x_pos
        cell_height = y_offs + height2 - y_pos
        # itemsize must conform to the cell space
        if not item_width or item_width > cell_width:
            item_width = cell_width
        if not item_height or item_height > cell_height:
            item_height = cell_height
        try:
            size_setter(
                item,
                pos=anchor(item_width, item_height, x_pos, y_pos, cell_width, cell_height),
                # BUG: DPG parses integers as unsigned longs, so a
                # negative value would actually make the item larger.
                width=max(int(item_width), 1),
                height=max(int(item_height), 1),
            )
        except SystemError:
            if not RegistryAPI.item_exists(item):
                if not del_items:
                    del_items = []
                del_items.append(item)
            else:
                raise
    if del_items:
        for item in del_items:
            grid.pop(item)




####################################################
############# GRID & GRID DEPENDENCIES #############
####################################################

_GT = TypeVar("_GT")
_ST = TypeVar("_ST")

class GridConfigure(Generic[_GT, _ST]):
    __slots__ = ("_get_eval", "_set_eval")

    def __set_name__(self, cls, name: str):
        self._get_eval = compile(f"instance._{name}", __name__, mode="eval")
        self._set_eval = compile(f"instance.configure({name}=value)", __name__, mode="exec")

    def __get__(self, instance, cls) -> _GT:
        if instance is None:
            return self  # type: ignore
        return eval(self._get_eval)

    def __set__(self, instance, value: _ST) -> None:
        exec(self._set_eval)


class GridMember(NamedTuple):
    item     : Item
    cell1    : tuple[int, int]
    cell2    : tuple[int, int]
    max_wt   : int
    max_ht   : int
    anchor   : FAnchor
    fset_size: Callable = _set_item_size


class GridOffset(NamedTuple):
    """Contains Grid padding or spacing values.

    Args:
        * x1: Horizontal offset from the left. Default is 0.

        * x1: Horizontal offset from the right. Default is 0.

        * y1: Vertical offset from the top. Default is 0.

        * y2: Vertical offset from the bottom. Default is 0.
    """
    x1: int = 0
    x2: int = 0
    y1: int = 0
    y2: int = 0

    @classmethod
    def new(cls, x: int | tuple[int, int], y: int | tuple[int, int], /) -> Self:
        """
        Args:
            * x: A single value to use as both x1 and x2 values, or a 2-tuple containing
            individual values for x1 and x2 respectively.

            * y: A single value to use as both y1 and y2 values, or a 2-tuple containing
            individual values for y1 and y2 respectively.
        """
        if isinstance(x, (int, float)):
            x = x, x
        if isinstance(y, (int, float)):
            y = y, y
        return cls(*x, *y)

    @property
    def x(self):
        return self[:2]

    @property
    def y(self):
        return self[2:]


class GridSlot:
    """A Grid row or column."""

    __slots__ = ("_weight", "_size", "label")

    FIXED = "fixed"
    SIZED = "sized"

    def __init__(self, label: str = ''):
        self.label   = label
        self._weight = 1.0
        self._size   = 0

    def __repr__(self):
        return f"{type(self).__qualname__}(label={self.label!r}, weight={self.weight}, size={self.size})"

    weight: GridConfigure[float, float] = GridConfigure()
    size  : GridConfigure[int, int]     = GridConfigure()

    @property
    def policy(self) -> Literal["fixed", "sized"]:
        return self.FIXED if self.size else self.SIZED

    @overload
    def configure(self, *, weight: float = ..., size: int = ...) -> None: ...  # type: ignore
    def configure(self, *, weight: Any = None, size: int | None = None) -> None:
        if weight is not None:
            self._weight = max(0, weight)
        if size is not None:
            self._size = max(0, size)


class GridAxis(Collection[GridSlot]):
    """A collection of Grid slots of a single axis."""

    __slots__ = ("_members", 'label',)

    _members: list[GridSlot]

    def __init__(self, length : int = 0, label: str = '') -> None:
        self.label    = label
        self._members = []
        self.resize(length)

    def __str__(self):
        return str(self._members)

    def __repr__(self):
        return f"{type(self).__qualname__}({str(self).strip('[]')})"

    def __len__(self):
        return len(self._members)

    def __int__(self):
        return len(self._members)

    def __iter__(self) -> Iterator[GridSlot]:
        yield from self._members

    def __getitem__(self, index: int) -> GridSlot:
        return self._members[index]

    def __contains__(self, other: Any) -> bool:
        return other in self._members

    def __add__(self, value: int) -> Self:
        self._members.extend(GridSlot(self.label) for _ in range(value))
        return self

    def __sub__(self, value: int) -> Self:
        if value > 0:
            del self._members[-value:]
        return self

    @property
    def weight(self) -> float:
        # ignore weight of fixed-size slots
        return sum(slot._weight for slot in self._members if not slot._size)

    @property
    def min_size(self) -> int:
        return sum(slot._size for slot in self._members)

    def insert(self, index: int, *args) -> None:
        """Insert a new row/column somewhere in the axis. The indexes of rows/columns
        trailing the point of insertion will be offset by 1.

        Args:
            * index: Specifies the position to insert the new row or column.
        """
        self._members.insert(index, GridSlot(self.label))

    def delete(self, index: int) -> None:
        self._members.pop(index)

    def resize(self, amount: int) -> None:
        """Update the size of the axis.

        Args:
            * amount: Integer specifying the amount of rows/columns to add or remove. A
            positive value will add that many additional rows/columns, while a negative
            value will remove that many from the end of the axis.
        """
        if amount >= 0:
            self._members.extend(GridSlot(self.label) for _ in range(amount))
        else:
            del self._members[-amount:]


class Grid:
    """A grid-layout manager for items.
        >>> import random
        >>> import dearpygui.dearpygui as dpg
        >>>
        >>>
        >>> def bind_button_theme(item: int | str):
        ...     rgb  = random.randint(0, 255), randint(0, 255), randint(0, 255)
        ...     with dpg.theme() as theme:
        ...         with dpg.theme_component(0):
        ...             dpg.add_theme_color(dpg.mvThemeCol_Button, [*rgb, 80])
        ...             dpg.add_theme_color(dpg.mvThemeCol_ButtonHovered, [*rgb, 255])
        ...     dpg.bind_item_theme(item, theme)
        >>>
        >>> def create_button() -> int | str:
        ...     item = dpg.add_button(label=tag, tag=tag)
        ...     bind_button_theme(item)
        ...     return item
        >>>
        >>>
        >>>    dpg.create_context()
        >>>    dpg.create_viewport(title="Grid Demo", width=1000, height=1000, min_height=10, min_width=10)
        >>>    dpg.setup_dearpygui()
        >>>
        >>>    with dpg.window(no_scrollbar=True, no_background=True) as win:
        ...        grid = Grid(6, 6, padding=((10, 30), 8), spacing=(0, 0))
        ...
        ...        # Without additional arguments, items will expand and shrink to the cell's size.
        ...        grid.push(create_button(),  0,  2)  # first row
        ...        grid.push(create_button(), -1,  3)  # last row
        ...
        ...        # You can clamp an item's width/height and include an alignment option. A few
        ...        # valid anchor options are 'n', 'ne', 'e', 'se', 's', 'sw', 'w', 'nw', and 'c'.
        ...        grid.push(create_button(), 1, 1, max_height=25, anchor="w")      # west anchor
        ...        grid.push(create_button(), 1, 1, max_width=25, anchor="north")   # north anchor
        ...        grid.push(create_button(), 4, 4, max_height=25, anchor="w")
        ...
        ...        # These items will occupy a range of cells.
        ...        grid.push(create_button(),  3,  1,  4,  2)
        ...        grid.push(create_button(),  1,  3,  2,  4)
        ...        grid.push(create_button(),  4,  0, -1,  1)
        ...        grid.push(create_button(),  0,  4,  1, -1)
        ...
        ...        for anchor in grid.ANCHORS:
        ...            kwargs = dict(max_width=50, max_height=50, anchor=anchor)
        ...            # pack to individual cells (cont.)
        ...            grid.push(create_button(),  0,  0, **kwargs)  # first row & column
        ...            grid.push(create_button(), -1, -1, **kwargs)  # last row & column
        ...            # pack to a "merged cell" (cont.)
        ...            grid.push(create_button(),  2,  2,  3,  3, **kwargs)
        ...
        ...        # You can change the weight or set a fixed size for a row/column.
        ...        grid.cols[2].weight = 0.5
        ...        grid.cols[3].weight = 0.5
        ...        grid.rows[2].weight = 0.5
        ...        grid.rows[3].weight = 0.5
        >>>
        >>> dpg.set_primary_window(win, True)
        >>> dpg.show_viewport()
        >>> dpg.set_viewport_resize_callback(grid)
        >>>
        >>> while dpg.is_dearpygui_running():
        ...     dpg.render_dearpygui_frame()
    """
    __slots__ = (
        "_rows",
        "_cols",
        "_padding",
        "_spacing",
        "_parent",
        "_sizer",
        "_members",
        "members",
    )

    def __init_subclass__(cls) -> None:
        # if __call__ is redefined, ensure that DearPyGui can still call it
        if "__call__" in cls.__dict__ and "__code__" not in cls.__dict__:
            cls.__dict__["__code__"] = cls.__call__.__code__

    def __init__(
        self,
        rows: int = 1,
        cols: int = 1,
        *,
        spacing: tuple[int | tuple[int, int], int | tuple[int, int]] = (0, 0),
        padding: tuple[int | tuple[int, int], int | tuple[int, int]] = (0, 0),
        parent : Item                                                =      0,
        sizer  : FGetSize                                            = size_from_rect,
    ):
        """Args:
            * rows: The initial number of virtual rows in the grid (min. 1).

            * cols: The initial number of virtual columns in the grid (min. 1).

            * spacing: Horizontal and vertical padding between cells.

            * padding: Horizontal and vertical padding between outer cells and the
            parent/target item's walls.

            * parent: Reference to an existing item or viewport. The width and
            height of the grid will scale to the parent's size.

            * sizer: A function that accepts *parent* as a positional argument
            and returns a 2-item sequence containing its' width and height.

        The first value of the *spacing* and *padding* arguments horizontal
        spacing/padding, while the second affects vertical spacing/padding. When
        the value is a number, it is used for both the left/upper-most and
        right/lower-most spacing (for horizontal and vertical respectively).
        Alternatively, the value can be a 2-item sequence of integers; the first
        integer will be used for the left/upper-most spacing while the second is
        used for the right/lower-most. For example, `spacing=(4, (2, 6))` will
        expand to `spacing=((4, 4), (2, 6))` automatically.

        If specified, the *parent* argument must be a reference to an existing
        item. Otherwise, it will automatically use the item on top of the
        container stack as the parent. You can also pass the viewport identifier,
        "DPG NOT USED YET", to specify the viewport. However, you must specify
        the correct sizer for it (i.e. *sizer=size_from_viewport*).

        """
        self._rows = GridAxis(rows, 'y')
        self._cols = GridAxis(cols, 'x')
        self._members: dict[Item, GridMember] = {}
        self.members = types.MappingProxyType(self._members)
        self.configure(spacing=spacing, padding=padding, parent=parent, sizer=sizer)

    __call__ = redraw_grid
    __code__ = __call__.__code__  # enables use as a direct DPG callback

    def __setitem__(self, coords: tuple[int, int] | tuple[int, int, int, int], item: Item, *, _anchor: str = "c"):
        """Pack and center an item into a grid cell or cell range."""
        self.push(item, *coords, anchor=_anchor)

    @property
    def rows(self) -> GridAxis:
        return self._rows
    @rows.setter
    def rows(self, value: int | GridAxis) -> None:
        if self._rows != value:  # self.rows += 1 -> self.rows.__set__(self, (self._rows + 1))
            self.configure(rows=value)  # type: ignore

    @property
    def cols(self) -> GridAxis:
        return self._cols
    @cols.setter
    def cols(self, value: int | GridAxis) -> None:
        if self._cols != value:  # self.cols += 1 -> self.cols.__set__(self, (self._cols + 1))
            self.configure(cols=value)  # type: ignore

    padding: GridConfigure[GridOffset, tuple[int | tuple[int, int], int | tuple[int, int]]] = GridConfigure()
    spacing: GridConfigure[GridOffset, tuple[int | tuple[int, int], int | tuple[int, int]]] = GridConfigure()

    parent : GridConfigure[Item  , Item  ]     = GridConfigure()
    sizer  : GridConfigure[FGetSize, FGetSize] = GridConfigure()

    ANCHORS = tuple(a.lower() for a in _ANCHOR_MAP)

    def configure(
        self,
        *,
        rows   : int = 0,
        cols   : int = 0,
        spacing: tuple[int | tuple[int, int], int | tuple[int, int]] | None = None,
        padding: tuple[int | tuple[int, int], int | tuple[int, int]] | None = None,
        parent : Item | None = None,
        sizer  : FGetSize | None = None,
    ):
        if rows > 0:
            self._rows.resize(rows - len(self.rows))
        if cols > 0:
            self._cols.resize(cols - len(self.cols))
        if spacing:
            self._spacing = GridOffset.new(*spacing)
        if padding:
            self._padding = GridOffset.new(*padding)
        if parent is not None:
            # emulate auto-parenting similar to DearPyGui
            if not parent and not (parent:=RegistryAPI.top_stack_item()):
                raise ValueError("requires parent (container stack is empty).")
            self._parent = parent
        if sizer:
            self._sizer = sizer

    @overload
    def push(self, item: Item, row: int, col: int, /, *, max_width: int = ..., max_height: int = ..., anchor: str = ..., normalize_indexes: bool = ...) -> None: ...
    @overload
    def push(self, item: Item, row_start: int, col_start: int, row_stop: int, col_stop: int, /, *, max_width: int = ..., max_height: int = ..., anchor: str = ..., normalize_indexes: bool = ...) -> None: ...
    def push(self, item: Item, r1: int, c1: int, r2: Any = None, c2: Any = None, /, *, max_width : int = 0, max_height: int = 0, anchor: str = "c", normalize_indexes: bool = False) -> None:
        """Attach an item to the grid. It will be managed until the item is deleted or
        is released from the grid via the `.pop` or `.clear` methods.

        Args:
            * item: Reference to an existing item.

            * row, row_start: The index of the row to place the item into. If *row_stop*
            is specified, this is the starting row index.

            * col, col_start: The index of the column to place the item into. If *col_stop*
            is specified, this is the starting column index.

            * row_stop: The index of the last row to place the item into.

            * col_stop: The index of the last column to place the item into.

            * max_width: Clamps the horizontal size (in pixels) of the item.

            * max_height: Clamps the vertical size (in pixels) of the item.

            * anchor: A case-insensitive string value specifying the cell wall(s) that
            will anchor the item in place. Appropriate values consist of any cardinal
            direction name or abbreviation ("north", "n", "southeast", "se", etc)
            and "center" (abbrv. as "c").

            * normalize_indexes: If True, converts any negative indexes to positive.
            False (default) preserves the original indexes.

        """
        # ensure item is currently valid
        if not RegistryAPI.item_exists(item):
            raise TypeError(f"`item` {item!r} does not exist.")
        # ensure coords are currently valid (XXX coords can become invalid as the number of slots change)
        try:
            self._rows[r1]
            self._rows[c1]
            self._cols[r2 or 0]
            self._cols[c2 or 0]
        except IndexError:
            raise IndexError("index(s) outside of grid range.") from None
        if r2 is None or c2 is None:
            # pushing to single cell
            r2, c2 = r1, c1
        if normalize_indexes:
            r1, c1, r2, c2 = normalize_cellspan(r1, c1, r2, c2, len(self._rows), len(self._cols))
        # get anchor calculator
        try:
            anchor_fn = _ANCHOR_MAP[anchor.lower()]
        except TypeError:
            raise TypeError(
                f"`anchor` expected as a str, (got {type(anchor)})."
            ) from None
        except KeyError:
            raise ValueError(
                f"invalid `anchor` argument {anchor!r}, try {', '.join(f'{a!r}' for a in self.ANCHORS)}."
            ) from None
        self._members[item] = GridMember(item, (r1, c1), (r2, c2), max(max_width, 0), max(max_height, 0), anchor_fn)

    def pop(self, item: Any):
        """Release an item from the grid."""
        return self._members.pop(item, None)

    def clear(self):
        """Release all items from the grid."""
        self._members.clear()

